import openmdao.api as om

from aviary.subsystems.mass.flops_based.empty_margin import EmptyMassMargin
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class MassSummation(om.Group):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        aviary_options: AviaryValues = self.options['aviary_options']
        alt_mass = aviary_options.get_val(Aircraft.Design.USE_ALT_MASS)

        self.add_subsystem(
            'structure_mass', StructureMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'propulsion_mass', PropulsionMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        if alt_mass:
            self.add_subsystem(
                'system_equip_mass_base',
                AltSystemsEquipMassBase(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'system_equip_mass',
                AltSystemsEquipMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

        else:
            self.add_subsystem(
                'system_equip_mass', SystemsEquipMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'empty_mass_margin', EmptyMassMargin(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        if alt_mass:
            self.add_subsystem(
                'empty_mass', AltEmptyMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

        else:
            self.add_subsystem('empty_mass', EmptyMass(aviary_options=aviary_options),
                               promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'operating_mass', OperatingMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'zero_fuel_mass', ZeroFuelMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('fuel_mass', FuelMass(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])


class StructureMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Canard.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Fins.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Fuselage.MASS, val=0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.MASS, val=0.0)
        add_aviary_input(self, Aircraft.LandingGear.MAIN_GEAR_MASS, val=0.0)
        add_aviary_input(self, Aircraft.LandingGear.NOSE_GEAR_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Nacelle.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Paint.MASS, val=0.0)
        add_aviary_input(self, Aircraft.VerticalTail.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.MASS, val=0.0)

        add_aviary_output(self, Aircraft.Design.STRUCTURE_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Design.STRUCTURE_MASS, '*', val=1)

    def compute(self, inputs, outputs):
        canard_mass = inputs[Aircraft.Canard.MASS]
        fin_mass = inputs[Aircraft.Fins.MASS]
        fus_mass = inputs[Aircraft.Fuselage.MASS]
        htail_mass = inputs[Aircraft.HorizontalTail.MASS]
        main_gear_mass = inputs[Aircraft.LandingGear.MAIN_GEAR_MASS]
        nose_gear_mass = inputs[Aircraft.LandingGear.NOSE_GEAR_MASS]
        nac_mass = inputs[Aircraft.Nacelle.MASS]
        paint_mass = inputs[Aircraft.Paint.MASS]
        vtail_mass = inputs[Aircraft.VerticalTail.MASS]
        wing_mass = inputs[Aircraft.Wing.MASS]

        outputs[Aircraft.Design.STRUCTURE_MASS] = \
            wing_mass + htail_mass + vtail_mass + fin_mass + canard_mass + \
            fus_mass + main_gear_mass + nose_gear_mass + nac_mass + paint_mass


class PropulsionMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_MISC_MASS, val=0.0)
        add_aviary_input(
            self, Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_ENGINE_MASS, val=0.0)

        add_aviary_output(self, Aircraft.Propulsion.MASS, val=0.0)

    def setup_partials(self):
        prop_wrt = [
            Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS,
            Aircraft.Propulsion.TOTAL_MISC_MASS,
            Aircraft.Fuel.FUEL_SYSTEM_MASS,
            Aircraft.Propulsion.TOTAL_ENGINE_MASS]

        self.declare_partials(Aircraft.Propulsion.MASS, prop_wrt, val=1)

    def compute(self, inputs, outputs):
        fuel_sys_mass = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS]
        misc_prop_mass = inputs[Aircraft.Propulsion.TOTAL_MISC_MASS]
        thrust_rev_mass = inputs[Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS]
        total_eng_mass = inputs[Aircraft.Propulsion.TOTAL_ENGINE_MASS]

        outputs[Aircraft.Propulsion.MASS] = \
            thrust_rev_mass + misc_prop_mass + fuel_sys_mass + total_eng_mass


class SystemsEquipMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.AirConditioning.MASS, val=0.0)
        add_aviary_input(self, Aircraft.AntiIcing.MASS, val=0.0)
        add_aviary_input(self, Aircraft.APU.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Avionics.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Electrical.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Furnishings.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Hydraulics.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Instruments.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.SURFACE_CONTROL_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS, val=0.0)

        add_aviary_output(self, Aircraft.Design.SYSTEMS_EQUIP_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Design.SYSTEMS_EQUIP_MASS, '*', val=1)

    def compute(self, inputs, outputs):
        AC_mass = inputs[Aircraft.AirConditioning.MASS]
        anti_icing_mass = inputs[Aircraft.AntiIcing.MASS]
        APU_mass = inputs[Aircraft.APU.MASS]
        avionics_mass = inputs[Aircraft.Avionics.MASS]
        elec_mass = inputs[Aircraft.Electrical.MASS]
        furnish_mass = inputs[Aircraft.Furnishings.MASS]
        hydraulics_mass = inputs[Aircraft.Hydraulics.MASS]
        instrument_mass = inputs[Aircraft.Instruments.MASS]
        surf_ctrl_mass = inputs[Aircraft.Wing.SURFACE_CONTROL_MASS]
        subsystems_mass = inputs[Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS]

        outputs[Aircraft.Design.SYSTEMS_EQUIP_MASS] = \
            surf_ctrl_mass + APU_mass + instrument_mass + hydraulics_mass + \
            elec_mass + avionics_mass + furnish_mass + AC_mass + anti_icing_mass + \
            subsystems_mass


class AltSystemsEquipMassBase(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.AirConditioning.MASS, val=0.0)
        add_aviary_input(self, Aircraft.AntiIcing.MASS, val=0.0)
        add_aviary_input(self, Aircraft.APU.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Avionics.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Electrical.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Furnishings.MASS_BASE, val=0.0)
        add_aviary_input(self, Aircraft.Hydraulics.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Instruments.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.SURFACE_CONTROL_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS, val=0.0)

        add_aviary_output(self, Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE, '*', val=1)

    def compute(self, inputs, outputs):
        AC_mass = inputs[Aircraft.AirConditioning.MASS]
        anti_icing_mass = inputs[Aircraft.AntiIcing.MASS]
        APU_mass = inputs[Aircraft.APU.MASS]
        avionics_mass = inputs[Aircraft.Avionics.MASS]
        elec_mass = inputs[Aircraft.Electrical.MASS]
        furnish_mass_base = inputs[Aircraft.Furnishings.MASS_BASE]
        hydraulics_mass = inputs[Aircraft.Hydraulics.MASS]
        instrument_mass = inputs[Aircraft.Instruments.MASS]
        surf_ctrl_mass = inputs[Aircraft.Wing.SURFACE_CONTROL_MASS]
        subsystems_mass = inputs[Aircraft.Design.EXTERNAL_SUBSYSTEMS_MASS]

        outputs[Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE] = \
            surf_ctrl_mass + APU_mass + instrument_mass + hydraulics_mass + \
            elec_mass + avionics_mass + furnish_mass_base + AC_mass + anti_icing_mass + \
            subsystems_mass


class AltSystemsEquipMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE, val=0.0)
        add_aviary_input(self, Aircraft.Design.STRUCTURE_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.MASS, val=0.0)

        add_aviary_output(self, Aircraft.Design.SYSTEMS_EQUIP_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(
            Aircraft.Design.SYSTEMS_EQUIP_MASS,
            [Aircraft.Design.STRUCTURE_MASS, Aircraft.Propulsion.MASS],
            val=0.01
        )

        self.declare_partials(
            Aircraft.Design.SYSTEMS_EQUIP_MASS,
            Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE,
            val=1.01
        )

    def compute(self, inputs, outputs):
        sys_equip_mass_base = inputs[Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE]
        structure_mass = inputs[Aircraft.Design.STRUCTURE_MASS]
        prop_mass = inputs[Aircraft.Propulsion.MASS]

        outputs[Aircraft.Design.SYSTEMS_EQUIP_MASS] = \
            sys_equip_mass_base + 0.01 * \
            (structure_mass + prop_mass + sys_equip_mass_base)


class EmptyMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Design.EMPTY_MASS_MARGIN, val=0.0)
        add_aviary_input(self, Aircraft.Design.STRUCTURE_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Design.SYSTEMS_EQUIP_MASS, val=0.0)

        add_aviary_output(self, Aircraft.Design.EMPTY_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Design.EMPTY_MASS, '*', val=1)

    def compute(self, inputs, outputs):
        structure_mass = inputs[Aircraft.Design.STRUCTURE_MASS]
        prop_mass = inputs[Aircraft.Propulsion.MASS]
        sys_equip_mass = inputs[Aircraft.Design.SYSTEMS_EQUIP_MASS]
        empty_mass_margin = inputs[Aircraft.Design.EMPTY_MASS_MARGIN]

        outputs[Aircraft.Design.EMPTY_MASS] = \
            structure_mass + prop_mass + sys_equip_mass + empty_mass_margin


class AltEmptyMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Design.EMPTY_MASS_MARGIN, val=0.0)
        add_aviary_input(self, Aircraft.Design.STRUCTURE_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.MASS, val=0.0)
        add_aviary_input(self, Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE, val=0.0)

        add_aviary_output(self, Aircraft.Design.EMPTY_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Design.EMPTY_MASS,
                              [Aircraft.Design.STRUCTURE_MASS,
                               Aircraft.Propulsion.MASS,
                               Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE],
                              val=1.01)
        self.declare_partials(Aircraft.Design.EMPTY_MASS,
                              Aircraft.Design.EMPTY_MASS_MARGIN,
                              val=1.0)

    def compute(self, inputs, outputs):
        structure_mass = inputs[Aircraft.Design.STRUCTURE_MASS]
        prop_mass = inputs[Aircraft.Propulsion.MASS]
        sys_equip_mass_base = inputs[Aircraft.Design.SYSTEMS_EQUIP_MASS_BASE]
        empty_mass_margin = inputs[Aircraft.Design.EMPTY_MASS_MARGIN]

        outputs[Aircraft.Design.EMPTY_MASS] = \
            1.01 * (structure_mass + prop_mass + sys_equip_mass_base) + empty_mass_margin


class OperatingMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.CrewPayload.CARGO_CONTAINER_MASS, val=0.0)
        add_aviary_input(self, Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS, val=0.0)
        add_aviary_input(self, Aircraft.CrewPayload.FLIGHT_CREW_MASS, val=0.0)
        add_aviary_input(self, Aircraft.CrewPayload.PASSENGER_SERVICE_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Design.EMPTY_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Fuel.UNUSABLE_FUEL_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_ENGINE_OIL_MASS, val=0.0)

        add_aviary_output(self, Aircraft.Design.OPERATING_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Design.OPERATING_MASS, '*', val=1)

    def compute(self, inputs, outputs):
        cargo_container_mass = inputs[Aircraft.CrewPayload.CARGO_CONTAINER_MASS]
        non_flight_crew_mass = inputs[Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS]
        flight_crew_mass = inputs[Aircraft.CrewPayload.FLIGHT_CREW_MASS]
        empty_mass = inputs[Aircraft.Design.EMPTY_MASS]
        oil_mass = inputs[Aircraft.Propulsion.TOTAL_ENGINE_OIL_MASS]
        pass_service_mass = inputs[Aircraft.CrewPayload.PASSENGER_SERVICE_MASS]
        unusable_fuel_mass = inputs[Aircraft.Fuel.UNUSABLE_FUEL_MASS]

        outputs[Aircraft.Design.OPERATING_MASS] = \
            empty_mass + non_flight_crew_mass + flight_crew_mass + \
            unusable_fuel_mass + oil_mass + pass_service_mass + cargo_container_mass


class ZeroFuelMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.CrewPayload.PASSENGER_MASS, val=0.0)
        add_aviary_input(self, Aircraft.CrewPayload.BAGGAGE_MASS, val=0.0)
        add_aviary_input(self, Aircraft.CrewPayload.CARGO_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Design.OPERATING_MASS, val=0.0)

        add_aviary_output(self, Aircraft.Design.ZERO_FUEL_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Aircraft.Design.ZERO_FUEL_MASS, '*', val=1)

    def compute(self, inputs, outputs):
        pass_mass = inputs[Aircraft.CrewPayload.PASSENGER_MASS]
        bag_mass = inputs[Aircraft.CrewPayload.BAGGAGE_MASS]
        cargo_mass = inputs[Aircraft.CrewPayload.CARGO_MASS]
        operating_mass = inputs[Aircraft.Design.OPERATING_MASS]

        outputs[Aircraft.Design.ZERO_FUEL_MASS] = \
            operating_mass + pass_mass + bag_mass + cargo_mass


class FuelMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Design.ZERO_FUEL_MASS, val=0.0)

        add_aviary_output(self, Mission.Design.FUEL_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(Mission.Design.FUEL_MASS,
                              Mission.Design.GROSS_MASS, val=1)
        self.declare_partials(Mission.Design.FUEL_MASS,
                              Aircraft.Design.ZERO_FUEL_MASS, val=-1)

    def compute(self, inputs, outputs):
        zero_fuel_mass = inputs[Aircraft.Design.ZERO_FUEL_MASS]
        gross_mass = inputs[Mission.Design.GROSS_MASS]

        outputs[Mission.Design.FUEL_MASS] = gross_mass - zero_fuel_mass



from copy import deepcopy
import unittest

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs

from aviary.interface.default_phase_info.two_dof import phase_info
from aviary.interface.methods_for_level1 import run_aviary
from aviary.variable_info.variables import Aircraft, Mission


@use_tempdirs
class ProblemPhaseTestCase(unittest.TestCase):

    @require_pyoptsparse(optimizer="IPOPT")
    def bench_test_swap_3_FwGm(self):
        local_phase_info = deepcopy(phase_info)
        prob = run_aviary('models/test_aircraft/aircraft_for_bench_FwGm.csv',
                          local_phase_info, optimizer='IPOPT')

        rtol = 1e-2

        # There are no truth values for these.
        assert_near_equal(prob.get_val(Mission.Design.GROSS_MASS),
                          186418., tolerance=rtol)

        assert_near_equal(prob.get_val(Aircraft.Design.OPERATING_MASS),
                          104530., tolerance=rtol)

        assert_near_equal(prob.get_val(Mission.Summary.TOTAL_FUEL_MASS),
                          44032., tolerance=rtol)

        assert_near_equal(prob.get_val('landing.' + Mission.Landing.GROUND_DISTANCE),
                          2528., tolerance=rtol)

        assert_near_equal(prob.get_val("traj.desc2.timeseries.distance")[-1],
                          3675.0, tolerance=rtol)

    @require_pyoptsparse(optimizer="SNOPT")
    def bench_test_swap_3_FwGm_SNOPT(self):
        local_phase_info = deepcopy(phase_info)
        prob = run_aviary('models/test_aircraft/aircraft_for_bench_FwGm.csv',
                          local_phase_info, optimizer='SNOPT')

        rtol = 1e-2

        # There are no truth values for these.
        assert_near_equal(prob.get_val(Mission.Design.GROSS_MASS),
                          186418., tolerance=rtol)

        assert_near_equal(prob.get_val(Aircraft.Design.OPERATING_MASS),
                          104530., tolerance=rtol)

        assert_near_equal(prob.get_val(Mission.Summary.TOTAL_FUEL_MASS),
                          44032., tolerance=rtol)

        assert_near_equal(prob.get_val('landing.' + Mission.Landing.GROUND_DISTANCE),
                          2528., tolerance=rtol)

        assert_near_equal(prob.get_val("traj.desc2.timeseries.distance")[-1],
                          3675.0, tolerance=rtol)


if __name__ == "__main__":
    # unittest.main()
    test = ProblemPhaseTestCase()
    test.bench_test_swap_3_FwGm_SNOPT()



import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.variable_info.enums import AnalysisScheme, AlphaModes, SpeedType
from aviary.variable_info.variables import Mission, Dynamic

from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.mission.gasp_based.ode.descent_eom import DescentRates
from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.constraints.flight_constraints import \
    FlightConstraints
from aviary.mission.gasp_based.ode.constraints.speed_constraints import SpeedConstraints
from aviary.mission.gasp_based.ode.descent_eom import DescentRates

from aviary.variable_info.enums import AnalysisScheme, SpeedType
from aviary.variable_info.variables import Dynamic
from aviary.subsystems.aerodynamics.aerodynamics_builder import AerodynamicsBuilderBase
from aviary.subsystems.propulsion.propulsion_builder import PropulsionBuilderBase
from aviary.mission.ode.specific_energy_rate import SpecificEnergyRate
from aviary.mission.ode.altitude_rate import AltitudeRate


class DescentODE(BaseODE):
    """ODE for quasi-steady descent.

    This ODE has a ``KSComp`` which allows for the switching of obeying an EAS
    constraint cruise Mach constraint, whichever is being violated. It is only included
    if the option to target the cruise speed is set to true, otherwise it is assumed
    that the equivalent airspeed is an input to this ODE and is set further up in the
    model.
    """

    def initialize(self):
        super().initialize()
        self.options.declare("input_speed_type", types=SpeedType,
                             desc="Whether the speed is given as a equivalent airspeed, true airspeed, or mach number")
        self.options.declare("alt_trigger_units")
        self.options.declare("speed_trigger_units")
        self.options.declare(
            "mach_cruise", default=0, desc="targeted cruise mach number"
        )
        self.options.declare(
            "EAS_limit", default=0, desc="maximum descending EAS in knots"
        )

    def setup(self):
        nn = self.options["num_nodes"]
        analysis_scheme = self.options["analysis_scheme"]
        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['core_subsystems']
        input_speed_type = self.options["input_speed_type"]

        flight_condition_group = self

        if input_speed_type is SpeedType.EAS:
            speed_inputs = ["EAS"]
            speed_outputs = ["mach", ("TAS", Dynamic.Mission.VELOCITY)]
        elif input_speed_type is SpeedType.MACH:
            speed_inputs = ["mach"]
            speed_outputs = ["EAS", ("TAS", Dynamic.Mission.VELOCITY)]

        # TODO: paramport
        self.add_subsystem("params", ParamPort(), promotes=["*"])

        self.add_subsystem(
            "USatm",
            USatm1976Comp(
                num_nodes=nn),
            promotes_inputs=[
                ("h",
                 Dynamic.Mission.ALTITUDE)],
            promotes_outputs=[
                "rho",
                ("sos",
                 Dynamic.Mission.SPEED_OF_SOUND),
                ("temp",
                 Dynamic.Mission.TEMPERATURE),
                ("pres",
                 Dynamic.Mission.STATIC_PRESSURE),
                "viscosity"],
        )

        if analysis_scheme is AnalysisScheme.COLLOCATION:
            EAS_limit = self.options["EAS_limit"]
            mach_cruise = self.options["mach_cruise"]
            constraint_args = {}
            integration_states = []
            constraint_inputs = []

            # Add a group to contain the balance

            if input_speed_type is SpeedType.MACH:
                mach_balance_group = self.add_subsystem(
                    "mach_balance_group", subsys=om.Group(), promotes=["*"]
                )

                mach_balance_group.nonlinear_solver = om.NewtonSolver()
                mach_balance_group.nonlinear_solver.options["solve_subsystems"] = True
                mach_balance_group.nonlinear_solver.options["iprint"] = 0
                mach_balance_group.nonlinear_solver.options["atol"] = 1e-7
                mach_balance_group.nonlinear_solver.options["rtol"] = 1e-7
                mach_balance_group.nonlinear_solver.linesearch = om.BoundsEnforceLS()
                mach_balance_group.linear_solver = om.DirectSolver(assemble_jac=True)

                speed_bal = om.BalanceComp(
                    name=Dynamic.Mission.MACH,
                    val=mach_cruise * np.ones(nn),
                    units="unitless",
                    lhs_name="KS",
                    rhs_val=0.0,
                    eq_units='unitless',
                    upper=1000,
                    lower=0,
                )
                mach_balance_group.add_subsystem(
                    "speed_bal",
                    speed_bal,
                    promotes_inputs=["KS"],
                    promotes_outputs=[Dynamic.Mission.MACH],
                )

                mach_balance_group.add_subsystem(
                    "speeds",
                    SpeedConstraints(
                        num_nodes=nn,
                        mach_cruise=mach_cruise,
                        EAS_target=EAS_limit,
                    ),
                    promotes_inputs=["EAS", Dynamic.Mission.MACH],
                    promotes_outputs=["speed_constraint"],
                )
                mach_balance_group.add_subsystem(
                    "ks",
                    om.KSComp(width=2, vec_size=nn, units="unitless"),
                    promotes_inputs=[("g", "speed_constraint")],
                    promotes_outputs=["KS"],
                )
                flight_condition_group = mach_balance_group

            lift_balance_group = self.add_subsystem(
                "lift_balance_group", subsys=om.Group(), promotes=["*"]
            )

        elif analysis_scheme is AnalysisScheme.SHOOTING:
            constraint_args = {'analysis_scheme': AnalysisScheme.SHOOTING,
                               'alt_trigger_units': self.options["alt_trigger_units"],
                               'speed_trigger_units': self.options["speed_trigger_units"]}
            integration_states = ["t_curr", Dynamic.Mission.DISTANCE]
            constraint_inputs = ["alt_trigger", "speed_trigger"]

            lift_balance_group = self

        flight_condition_group.add_subsystem(
            "fc",
            FlightConditions(num_nodes=nn, input_speed_type=input_speed_type),
            promotes_inputs=["rho", Dynamic.Mission.SPEED_OF_SOUND] + speed_inputs,
            promotes_outputs=[Dynamic.Mission.DYNAMIC_PRESSURE,] + speed_outputs,
        )

        # maybe replace this with the solver in AddAlphaControl?
        lift_balance_group.nonlinear_solver = om.NewtonSolver()
        lift_balance_group.nonlinear_solver.options["solve_subsystems"] = True
        lift_balance_group.nonlinear_solver.options["iprint"] = 0
        lift_balance_group.nonlinear_solver.options["atol"] = 1e-7
        lift_balance_group.nonlinear_solver.options["rtol"] = 1e-7
        lift_balance_group.nonlinear_solver.linesearch = om.BoundsEnforceLS()
        lift_balance_group.linear_solver = om.DirectSolver(assemble_jac=True)

        lift_balance_group.add_subsystem(
            "eom",
            DescentRates(
                num_nodes=nn,
                analysis_scheme=analysis_scheme),
            promotes_inputs=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.VELOCITY,
                Dynamic.Mission.DRAG,
                Dynamic.Mission.THRUST_TOTAL,
                "alpha",] +
            integration_states,
            promotes_outputs=[
                Dynamic.Mission.ALTITUDE_RATE,
                Dynamic.Mission.DISTANCE_RATE,
                "required_lift",
                Dynamic.Mission.FLIGHT_PATH_ANGLE,
            ],
        )

        self.add_subsystem(
            "constraints",
            FlightConstraints(num_nodes=nn, **constraint_args),
            promotes_inputs=[
                Dynamic.Mission.MASS,
                "alpha",
                "rho",
                "CL_max",
                Dynamic.Mission.FLIGHT_PATH_ANGLE,
                ("TAS", Dynamic.Mission.VELOCITY),
            ]
            + ["aircraft:*"]
            + constraint_inputs,
            promotes_outputs=["theta", "TAS_violation"],
        )

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options,
                  'method': 'cruise'}
        # collect the propulsion group names for later use
        prop_groups = []
        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                if isinstance(subsystem, AerodynamicsBuilderBase):
                    lift_balance_group.add_subsystem(subsystem.name,
                                                     system,
                                                     promotes_inputs=subsystem.mission_inputs(
                                                         **kwargs),
                                                     promotes_outputs=subsystem.mission_outputs(**kwargs))
                else:
                    if isinstance(subsystem, PropulsionBuilderBase):
                        prop_groups.append(subsystem.name)

                    self.add_subsystem(subsystem.name,
                                       system,
                                       promotes_inputs=subsystem.mission_inputs(
                                           **kwargs),
                                       promotes_outputs=subsystem.mission_outputs(**kwargs))

        self.AddAlphaControl(
            alpha_group=lift_balance_group,
            alpha_mode=AlphaModes.REQUIRED_LIFT,
            add_default_solver=False,
            num_nodes=nn)

        # the last two subsystems will also be used for constraints
        self.add_excess_rate_comps(nn)

        if analysis_scheme is AnalysisScheme.COLLOCATION:
            fc_loc = ['fc']
            if input_speed_type is SpeedType.MACH:
                mach_balance_group.set_order(['speed_bal', 'fc', 'speeds', 'ks'])
                fc_loc = ['mach_balance_group']
            # TODO this assumes the name of propulsion subsystem, need to pull from
            #      the subsystem itself
            self.set_order(['params',
                            'USatm',] +
                           fc_loc+prop_groups +
                           ['lift_balance_group',
                            'constraints',
                            'SPECIFIC_ENERGY_RATE_EXCESS',
                            'ALTITUDE_RATE_MAX'])

        ParamPort.set_default_vals(self)
        self.set_input_defaults(Dynamic.Mission.ALTITUDE,
                                val=37500 * np.ones(nn), units="ft")
        self.set_input_defaults(Dynamic.Mission.MASS,
                                val=147000 * np.ones(nn), units="lbm")
        self.set_input_defaults(Dynamic.Mission.MACH,
                                val=0 * np.ones(nn), units="unitless")
        self.set_input_defaults(Dynamic.Mission.THROTTLE,
                                val=0 * np.ones(nn), units="unitless")



import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp
from aviary.subsystems.mass.mass_to_weight import MassToWeight

from aviary.variable_info.enums import AlphaModes, AnalysisScheme, SpeedType
from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.flight_path_eom import FlightPathEOM
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.subsystems.propulsion.propulsion_builder import PropulsionBuilderBase
from aviary.variable_info.variables import Aircraft, Dynamic, Mission
from aviary.mission.ode.specific_energy_rate import SpecificEnergyRate
from aviary.mission.ode.altitude_rate import AltitudeRate


class FlightPathODE(BaseODE):
    """ ODE using 2D aircraft equations of motion with states distance, alt, TAS, and gamma.

    Control is managed via angle-of-attack (alpha).
    """

    def initialize(self):
        super().initialize()
        self.options.declare("alpha_mode", default=AlphaModes.DEFAULT, types=AlphaModes)
        self.options.declare("input_speed_type", default=SpeedType.TAS, types=SpeedType,
                             desc="Whether the speed is given as a equivalent airspeed, true airspeed, or mach number")
        self.options.declare(
            "ground_roll",
            types=bool,
            default=False,
            desc="True if the aircraft is confined to the ground. Removes altitude rate as an "
            "output and adjusts the TAS rate equation.")
        self.options.declare(
            "clean",
            types=bool,
            default=False,
            desc="If true then no flaps or gear are included. Useful for high-speed flight phases.")

    def setup(self):
        nn = self.options["num_nodes"]
        aviary_options = self.options['aviary_options']
        alpha_mode = self.options['alpha_mode']
        analysis_scheme = self.options['analysis_scheme']
        input_speed_type = self.options['input_speed_type']

        print_level = 0

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options,
                  'method': 'low_speed'}
        if self.options['clean']:
            kwargs['method'] = 'cruise'
            kwargs['output_alpha'] = False

        if input_speed_type is SpeedType.EAS:
            speed_inputs = ["EAS"]
            speed_outputs = ["mach", ("TAS", Dynamic.Mission.VELOCITY)]
        elif input_speed_type is SpeedType.TAS:
            speed_inputs = [("TAS", Dynamic.Mission.VELOCITY)]
            speed_outputs = ["mach", "EAS"]
        elif input_speed_type is SpeedType.MACH:
            speed_inputs = ["mach"]
            speed_outputs = ["EAS", ("TAS", Dynamic.Mission.VELOCITY)]

        EOM_inputs = [
            Dynamic.Mission.MASS,
            Dynamic.Mission.THRUST_TOTAL,
            Dynamic.Mission.LIFT,
            Dynamic.Mission.DRAG,
            Dynamic.Mission.VELOCITY,
            Dynamic.Mission.FLIGHT_PATH_ANGLE,
        ] + ['aircraft:*']
        if not self.options['ground_roll']:
            EOM_inputs.append('alpha')

        if analysis_scheme is AnalysisScheme.SHOOTING:
            EOM_inputs.append(Dynamic.Mission.ALTITUDE)
            EOM_inputs.append('distance_trigger')
            EOM_inputs.append('distance')
            EOM_inputs.append('t_curr')

            prop_group = om.Group()
        else:
            prop_group = self
        core_subsystems = self.options['core_subsystems']

        # TODO: paramport
        flight_path_params = ParamPort()
        if analysis_scheme is AnalysisScheme.SHOOTING and kwargs['method'] == 'cruise':
            flight_path_params.add_params({
                Aircraft.Design.OPERATING_MASS: dict(units='lbm', val=0),
                Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS: dict(units='lbm', val=0),
                Mission.Design.RESERVE_FUEL: dict(units='lbm', val=0),
            })
        self.add_subsystem("params", flight_path_params, promotes=["*"])

        self.add_subsystem(
            "USatm",
            USatm1976Comp(num_nodes=nn),
            promotes_inputs=[("h", Dynamic.Mission.ALTITUDE)],
            promotes_outputs=["rho", ("sos", Dynamic.Mission.SPEED_OF_SOUND),
                              ("temp", Dynamic.Mission.TEMPERATURE), ("pres", Dynamic.Mission.STATIC_PRESSURE), "viscosity", "drhos_dh"],
        )

        self.add_subsystem(
            "fc",
            FlightConditions(num_nodes=nn, input_speed_type=input_speed_type),
            promotes_inputs=["rho", Dynamic.Mission.SPEED_OF_SOUND] + speed_inputs,
            promotes_outputs=[Dynamic.Mission.DYNAMIC_PRESSURE,] + speed_outputs,
        )

        lift_comp = []
        if alpha_mode is AlphaModes.DEFAULT:
            # alpha as input
            pass
        else:
            if alpha_mode is AlphaModes.REQUIRED_LIFT:
                self.add_subsystem(
                    "calc_weight",
                    MassToWeight(num_nodes=nn),
                    promotes_inputs=[("mass", Dynamic.Mission.MASS)],
                    promotes_outputs=["weight"]
                )
                self.add_subsystem(
                    'calc_lift',
                    om.ExecComp(
                        'required_lift = weight*cos(alpha + gamma) - thrust*sin(i_wing)',
                        required_lift={'val': 0, 'units': 'lbf'},
                        weight={'val': 0, 'units': 'lbf'},
                        thrust={'val': 0, 'units': 'lbf'},
                        alpha={'val': 0, 'units': 'rad'},
                        gamma={'val': 0, 'units': 'rad'},
                        i_wing={'val': 0, 'units': 'rad'},
                    ),
                    promotes_inputs=[
                        'weight',
                        ('thrust', Dynamic.Mission.THRUST_TOTAL),
                        'alpha',
                        ('gamma', Dynamic.Mission.FLIGHT_PATH_ANGLE),
                        ('i_wing', Aircraft.Wing.INCIDENCE)
                    ],
                    promotes_outputs=['required_lift']
                )
                lift_comp = ['calc_weight', 'calc_lift']
            self.AddAlphaControl(alpha_mode=alpha_mode, target_load_factor=1,
                                 atol=1e-6, rtol=1e-12, num_nodes=nn, print_level=print_level)

        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                if isinstance(subsystem, PropulsionBuilderBase):
                    prop_group.add_subsystem(subsystem.name,
                                             system,
                                             promotes_inputs=subsystem.mission_inputs(
                                                 **kwargs),
                                             promotes_outputs=subsystem.mission_outputs(**kwargs))
                else:
                    self.add_subsystem(subsystem.name,
                                       system,
                                       promotes_inputs=subsystem.mission_inputs(
                                           **kwargs),
                                       promotes_outputs=subsystem.mission_outputs(**kwargs))

        if analysis_scheme is AnalysisScheme.SHOOTING:
            prop_group.add_subsystem(
                'calc_thrust',
                om.ExecComp(
                    # TODO fix engines not providing thrust in the right direction
                    # + weight*sin(alpha + gamma)
                    'required_thrust = ( drag  ) / cos(i_wing)',
                    required_thrust={'val': 0, 'units': 'lbf'},
                    drag={'val': 0, 'units': 'lbf'},
                    # weight={'val': 0, 'units': 'lbf'},
                    # alpha={'val': 0, 'units': 'rad'},
                    # gamma={'val': 0, 'units': 'rad'},
                    i_wing={'val': 0, 'units': 'rad'},
                ),
                promotes_inputs=[
                    ('drag', Dynamic.Mission.DRAG),
                    # 'weight',
                    # 'alpha',
                    # ('gamma', Dynamic.Mission.FLIGHT_PATH_ANGLE),
                    ('i_wing', Aircraft.Wing.INCIDENCE)
                ],
                promotes_outputs=['required_thrust']
            )

            self.AddThrottleControl(prop_group=prop_group,
                                    atol=1e-8, print_level=print_level)

        self.add_subsystem(
            "eoms",
            FlightPathEOM(
                num_nodes=nn,
                ground_roll=self.options['ground_roll'],
                analysis_scheme=analysis_scheme),
            promotes_inputs=EOM_inputs,
            promotes_outputs=[
                Dynamic.Mission.VELOCITY_RATE,
                Dynamic.Mission.DISTANCE_RATE,
                "normal_force",
                "fuselage_pitch",
                "load_factor",
            ],
        )

        if not self.options['ground_roll']:
            self.promotes('eoms', outputs=[
                          Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE])

        self.add_excess_rate_comps(nn)

        # Example of how to use a print_comp
        debug_comp = []
        if False:
            from aviary.utils.functions import create_printcomp
            dummy_comp = create_printcomp(
                all_inputs=[
                    Aircraft.Design.OPERATING_MASS,
                    Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS,  # should be total
                    Mission.Design.RESERVE_FUEL,
                    Dynamic.Mission.DISTANCE,
                    Dynamic.Mission.THROTTLE,
                    'required_lift',
                    'load_factor',
                    Dynamic.Mission.ALTITUDE,
                    Dynamic.Mission.FLIGHT_PATH_ANGLE,
                ],
                input_units={
                    'required_lift': 'lbf',
                    Dynamic.Mission.FLIGHT_PATH_ANGLE: 'deg',
                })
            self.add_subsystem(
                "dummy_comp",
                dummy_comp(),
                promotes_inputs=["*"],)
            self.set_input_defaults(
                Dynamic.Mission.DISTANCE, val=0, units='NM')
            debug_comp = ['dummy_comp']

        if analysis_scheme is AnalysisScheme.SHOOTING:
            self.add_subsystem('mass_trigger',
                               om.ExecComp(
                                   'mass_trigger = OEM + payload + reserve_fuel + descent_fuel',
                                   mass_trigger={'val': 0, 'units': 'lbm'},
                                   OEM={'val': 0, 'units': 'lbm'},
                                   payload={'val': 0, 'units': 'lbm'},
                                   reserve_fuel={'val': 0, 'units': 'lbm'},
                                   descent_fuel={'val': 0, 'units': 'lbm'},
                               ),
                               promotes_inputs=[
                                   ('OEM', Aircraft.Design.OPERATING_MASS),
                                   ('payload', Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS),
                                   ('reserve_fuel', Mission.Design.RESERVE_FUEL),
                                   'descent_fuel',
                               ],
                               )

            self.set_order(['params', 'USatm', 'fc',
                            ] + lift_comp + [
                'core_aerodynamics',
                'alpha_comp',
                'prop_group',
                'eoms',
                'mass_trigger',
                'SPECIFIC_ENERGY_RATE_EXCESS',
                'ALTITUDE_RATE_MAX',
            ] +
                debug_comp)

        ParamPort.set_default_vals(self)
        if not self.options["clean"]:
            self.set_input_defaults("t_init_flaps", val=47.5)
            self.set_input_defaults("t_init_gear", val=37.3)
            self.set_input_defaults("t_curr", val=np.zeros(nn), units="s")
        self.set_input_defaults("alpha", val=np.zeros(nn), units="rad")
        self.set_input_defaults(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                                val=np.zeros(nn), units="deg")
        self.set_input_defaults(Dynamic.Mission.ALTITUDE, val=np.zeros(nn), units="ft")
        self.set_input_defaults(Dynamic.Mission.MACH, val=np.zeros(nn), units="unitless")
        self.set_input_defaults(Dynamic.Mission.MASS, val=np.zeros(nn), units="lbm")
        self.set_input_defaults(Dynamic.Mission.VELOCITY, val=np.zeros(nn), units="kn")



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class WingMassSolve(om.ImplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400)
        add_aviary_input(self, Aircraft.Wing.HIGH_LIFT_MASS, val=3645)
        self.add_input(
            "c_strut_braced",
            val=1.00000001,
            units="unitless",
            desc="SKSTR: reduction in bending moment factor for strut braced wing",
        )
        add_aviary_input(self, Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.75)
        add_aviary_input(self, Aircraft.Wing.MASS_COEFFICIENT, val=102.5)
        add_aviary_input(self, Aircraft.Wing.MATERIAL_FACTOR, val=1.22130632)
        add_aviary_input(self, Aircraft.Engine.POSITION_FACTOR, val=0.98)
        self.add_input(
            "c_gear_loc",
            val=1.000000001,
            units="unitless",
            desc="SKGEAR: landing gear location factor",
        )
        add_aviary_input(self, Aircraft.Wing.SPAN, val=117.8)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.33)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15)
        self.add_input(
            "half_sweep",
            val=0.3947081519,
            units="rad",
            desc="SWC2: wing half-chord sweep angle",
        )

        self.add_output(
            "isolated_wing_mass",
            val=17670,
            units="lbm",
            desc="WW: wing mass including high lift devices (but excluding struts and fold effects)",
        )

        self.declare_partials("isolated_wing_mass", "*")

    def apply_nonlinear(self, inputs, outputs, residuals):

        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        high_lift_wt = inputs[Aircraft.Wing.HIGH_LIFT_MASS] * GRAV_ENGLISH_LBM
        c_strut_braced = inputs["c_strut_braced"]
        ULF = inputs[Aircraft.Wing.ULTIMATE_LOAD_FACTOR]
        c_wing_mass = inputs[Aircraft.Wing.MASS_COEFFICIENT]
        c_material = inputs[Aircraft.Wing.MATERIAL_FACTOR]
        c_eng_pos = inputs[Aircraft.Engine.POSITION_FACTOR]
        c_gear_loc = inputs["c_gear_loc"]
        wingspan = inputs[Aircraft.Wing.SPAN]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        tc_ratio_root = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_ROOT]
        half_sweep = inputs["half_sweep"]

        isolated_wing_wt = outputs["isolated_wing_mass"] * GRAV_ENGLISH_LBM

        foo = (
            c_strut_braced * ULF * (gross_wt_initial - 0.8 * isolated_wing_wt)
        ) ** 0.757
        wing_wt_guess = (
            c_wing_mass
            * c_material
            * c_eng_pos
            * c_gear_loc
            * foo
            * wingspan**1.049
            * (1.0 + taper_ratio) ** 0.4
        ) / (
            100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535
        ) + high_lift_wt

        residuals["isolated_wing_mass"] = (
            isolated_wing_wt - wing_wt_guess) / GRAV_ENGLISH_LBM

    def linearize(self, inputs, outputs, J):

        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        high_lift_wt = inputs[Aircraft.Wing.HIGH_LIFT_MASS] * GRAV_ENGLISH_LBM
        c_strut_braced = inputs["c_strut_braced"]
        ULF = inputs[Aircraft.Wing.ULTIMATE_LOAD_FACTOR]
        c_wing_mass = inputs[Aircraft.Wing.MASS_COEFFICIENT]
        c_material = inputs[Aircraft.Wing.MATERIAL_FACTOR]
        c_eng_pos = inputs[Aircraft.Engine.POSITION_FACTOR]
        c_gear_loc = inputs["c_gear_loc"]
        wingspan = inputs[Aircraft.Wing.SPAN]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        tc_ratio_root = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_ROOT]
        half_sweep = inputs["half_sweep"]

        isolated_wing_wt = outputs["isolated_wing_mass"] * GRAV_ENGLISH_LBM

        foo = (
            c_strut_braced * ULF * (gross_wt_initial - 0.8 * isolated_wing_wt)
        ) ** 0.757
        wing_wt_guess = (
            c_wing_mass
            * c_material
            * c_eng_pos
            * c_gear_loc
            * foo
            * wingspan**1.049
            * (1.0 + taper_ratio) ** 0.4
        ) / (
            100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535
        ) + high_lift_wt

        J["isolated_wing_mass", Mission.Design.GROSS_MASS] = (
            -(
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * wingspan**1.049
                * (1.0 + taper_ratio) ** 0.4
            )
            / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
            * 0.757
            * (c_strut_braced * ULF * (gross_wt_initial - 0.8 * isolated_wing_wt))
            ** (-0.243)
            * c_strut_braced
            * ULF
        )
        J["isolated_wing_mass", Aircraft.Wing.HIGH_LIFT_MASS] = -1
        J["isolated_wing_mass", "c_strut_braced"] = (
            -(
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * wingspan**1.049
                * (1.0 + taper_ratio) ** 0.4
            )
            / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
            * 0.757
            * (c_strut_braced * ULF * (gross_wt_initial - 0.8 * isolated_wing_wt))
            ** (-0.243)
            * ULF
            * (gross_wt_initial - 0.8 * isolated_wing_wt)
            / GRAV_ENGLISH_LBM
        )
        J["isolated_wing_mass", Aircraft.Wing.ULTIMATE_LOAD_FACTOR] = (
            -(
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * wingspan**1.049
                * (1.0 + taper_ratio) ** 0.4
            )
            / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
            * 0.757
            * (c_strut_braced * ULF * (gross_wt_initial - 0.8 * isolated_wing_wt))
            ** (-0.243)
            * c_strut_braced
            * (gross_wt_initial - 0.8 * isolated_wing_wt)
            / GRAV_ENGLISH_LBM
        )
        J["isolated_wing_mass", Aircraft.Wing.MASS_COEFFICIENT] = -(
            c_material
            * c_eng_pos
            * c_gear_loc
            * foo
            * wingspan**1.049
            * (1.0 + taper_ratio) ** 0.4
            / GRAV_ENGLISH_LBM
        ) / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
        J["isolated_wing_mass", Aircraft.Wing.MATERIAL_FACTOR] = -(
            c_wing_mass
            * c_eng_pos
            * c_gear_loc
            * foo
            * wingspan**1.049
            * (1.0 + taper_ratio) ** 0.4
            / GRAV_ENGLISH_LBM
        ) / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
        J["isolated_wing_mass", Aircraft.Engine.POSITION_FACTOR] = -(
            c_wing_mass
            * c_material
            * c_gear_loc
            * foo
            * wingspan**1.049
            * (1.0 + taper_ratio) ** 0.4
            / GRAV_ENGLISH_LBM
        ) / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
        J["isolated_wing_mass", "c_gear_loc"] = -(
            c_wing_mass
            * c_material
            * c_eng_pos
            * foo
            * wingspan**1.049
            * (1.0 + taper_ratio) ** 0.4
            / GRAV_ENGLISH_LBM
        ) / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
        J["isolated_wing_mass", Aircraft.Wing.SPAN] = (
            -1.049
            * (
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * foo
                * wingspan**0.049
                * (1.0 + taper_ratio) ** 0.4
            )
            / GRAV_ENGLISH_LBM
            / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
        )
        J["isolated_wing_mass", Aircraft.Wing.TAPER_RATIO] = (
            -0.4
            * (
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * foo
                * wingspan**1.049
                * (1.0 + taper_ratio) ** (-0.6)
            )
            / GRAV_ENGLISH_LBM
            / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
        )
        J["isolated_wing_mass", Aircraft.Wing.THICKNESS_TO_CHORD_ROOT] = (
            0.4
            * (
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * foo
                * wingspan**1.049
                * (1.0 + taper_ratio) ** 0.4
            )
            / GRAV_ENGLISH_LBM
            / (100000.0 * tc_ratio_root**1.4 * np.cos(half_sweep) ** 1.535)
        )
        J["isolated_wing_mass", "half_sweep"] = (
            1.535
            * (
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * foo
                * wingspan**1.049
                * (1.0 + taper_ratio) ** 0.4
            )
            / GRAV_ENGLISH_LBM
            / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 2.535)
            * (-np.sin(half_sweep))
        )
        J["isolated_wing_mass", "isolated_wing_mass"] = (
            1
            - (
                c_wing_mass
                * c_material
                * c_eng_pos
                * c_gear_loc
                * wingspan**1.049
                * (1.0 + taper_ratio) ** 0.4
            )
            / (100000.0 * tc_ratio_root**0.4 * np.cos(half_sweep) ** 1.535)
            * 0.757
            * (c_strut_braced * ULF * (gross_wt_initial - 0.8 * isolated_wing_wt))
            ** (-0.243)
            * (-0.8)
            * c_strut_braced
            * ULF
        )


class WingMassTotal(om.ExplicitComponent):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        self.add_input(
            "isolated_wing_mass",
            val=1500,
            units="lbm",
            desc="WW: wing mass including high lift devices (but excluding struts and fold effects)",
        )

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            add_aviary_input(self, Aircraft.Strut.MASS_COEFFICIENT, val=0.000000000001)

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless') == True:

            add_aviary_input(self, Aircraft.Wing.AREA, val=100)
            add_aviary_input(self, Aircraft.Wing.FOLDING_AREA, val=50)
            add_aviary_input(self, Aircraft.Wing.FOLD_MASS_COEFFICIENT, val=0.2)

        add_aviary_output(self, Aircraft.Wing.MASS, val=0)
        add_aviary_output(self, Aircraft.Strut.MASS, val=0)
        add_aviary_output(self, Aircraft.Wing.FOLD_MASS, val=0)

        self.declare_partials(Aircraft.Wing.MASS, "*")
        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            self.declare_partials(Aircraft.Strut.MASS, [
                                  Aircraft.Strut.MASS_COEFFICIENT, "isolated_wing_mass"])
        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):
            self.declare_partials(Aircraft.Wing.FOLD_MASS, [
                                  Aircraft.Wing.AREA, Aircraft.Wing.FOLDING_AREA, Aircraft.Wing.FOLD_MASS_COEFFICIENT, "isolated_wing_mass"])

    def compute(self, inputs, outputs):

        isolated_wing_wt = inputs["isolated_wing_mass"] * GRAV_ENGLISH_LBM

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            c_strut_mass = inputs[Aircraft.Strut.MASS_COEFFICIENT]

            strut_wt = c_strut_mass * isolated_wing_wt
            outputs[Aircraft.Strut.MASS] = strut_wt / GRAV_ENGLISH_LBM

        else:
            outputs[Aircraft.Strut.MASS] = strut_wt = 0

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):
            wing_area = inputs[Aircraft.Wing.AREA]
            folding_area = inputs[Aircraft.Wing.FOLDING_AREA]
            c_wing_fold = inputs[Aircraft.Wing.FOLD_MASS_COEFFICIENT]

            wt_per_area = isolated_wing_wt / wing_area
            temp_fold_wt = folding_area * wt_per_area
            fold_wt = c_wing_fold * temp_fold_wt
            outputs[Aircraft.Wing.FOLD_MASS] = fold_wt / GRAV_ENGLISH_LBM

        else:
            outputs[Aircraft.Wing.FOLD_MASS] = fold_wt = 0

        total_wing_wt = isolated_wing_wt + strut_wt + fold_wt
        outputs[Aircraft.Wing.MASS] = total_wing_wt / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):

        isolated_wing_wt = inputs["isolated_wing_mass"] * GRAV_ENGLISH_LBM

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            c_strut_mass = inputs[Aircraft.Strut.MASS_COEFFICIENT]

            J[Aircraft.Wing.MASS, Aircraft.Strut.MASS_COEFFICIENT] = \
                J[Aircraft.Strut.MASS, Aircraft.Strut.MASS_COEFFICIENT] = \
                isolated_wing_wt / GRAV_ENGLISH_LBM
            J[Aircraft.Wing.MASS, "isolated_wing_mass"] = 1 + c_strut_mass
            J[Aircraft.Strut.MASS, "isolated_wing_mass"] = c_strut_mass

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):
            wing_area = inputs[Aircraft.Wing.AREA]
            folding_area = inputs[Aircraft.Wing.FOLDING_AREA]
            c_wing_fold = inputs[Aircraft.Wing.FOLD_MASS_COEFFICIENT]

            wt_per_area = isolated_wing_wt / wing_area
            temp_fold_wt = folding_area * wt_per_area

            J[Aircraft.Wing.MASS, Aircraft.Wing.AREA] = \
                J[Aircraft.Wing.FOLD_MASS, Aircraft.Wing.AREA] = (
                -c_wing_fold * folding_area * isolated_wing_wt / wing_area**2 / GRAV_ENGLISH_LBM
            )
            J[Aircraft.Wing.MASS, Aircraft.Wing.FOLDING_AREA] = \
                J[Aircraft.Wing.FOLD_MASS, Aircraft.Wing.FOLDING_AREA] = \
                c_wing_fold * wt_per_area / GRAV_ENGLISH_LBM
            J[Aircraft.Wing.MASS, Aircraft.Wing.FOLD_MASS_COEFFICIENT] = \
                J[Aircraft.Wing.FOLD_MASS, Aircraft.Wing.FOLD_MASS_COEFFICIENT] = \
                temp_fold_wt / GRAV_ENGLISH_LBM
            J[Aircraft.Wing.MASS, "isolated_wing_mass"] = (
                1 + c_wing_fold * folding_area / wing_area
            )
            J[Aircraft.Wing.FOLD_MASS, "isolated_wing_mass"] = c_wing_fold * \
                folding_area / wing_area

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless') and \
                self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):

            J[Aircraft.Wing.MASS, "isolated_wing_mass"] = (
                1 + c_wing_fold * folding_area / wing_area + c_strut_mass
            )

        if (
            self.options["aviary_options"].get_val(
                Aircraft.Wing.HAS_STRUT, units='unitless') == False
            and self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless') == False
        ):
            J[Aircraft.Wing.MASS, "isolated_wing_mass"] = 1


class WingMassGroup(om.Group):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        # variables that are calculated at a higher level
        higher_level_inputs_isolated = [
            "c_strut_braced",
            "c_gear_loc",
            "half_sweep",
        ]
        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless') or \
                self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):

            higher_level_inputs_total = [
                "aircraft:*"
            ]
        else:
            higher_level_inputs_total = []

        # variables that are passed within the group but not used at a higher level
        connected_outputs_isolated = [
            "isolated_wing_mass",
        ]
        connected_inputs_total = [
            "isolated_wing_mass",
        ]

        isolated_mass = self.add_subsystem(
            "isolated_mass",
            WingMassSolve(aviary_options=aviary_options),
            promotes_inputs=higher_level_inputs_isolated + ["aircraft:*", "mission:*"],
            promotes_outputs=connected_outputs_isolated,
        )

        total_mass = self.add_subsystem(
            "total_mass",
            WingMassTotal(
                aviary_options=aviary_options,
            ),
            promotes_inputs=connected_inputs_total + higher_level_inputs_total,
            promotes_outputs=["aircraft:*"],
        )

        newton = isolated_mass.nonlinear_solver = om.NewtonSolver()

        newton.options["atol"] = 1e-9
        newton.options["rtol"] = 1e-9
        newton.options["iprint"] = 2
        newton.options["maxiter"] = 10
        newton.options["solve_subsystems"] = True
        newton.options["max_sub_solves"] = 10
        newton.options["err_on_non_converge"] = True
        newton.options["reraise_child_analysiserror"] = False
        newton.linesearch = om.BoundsEnforceLS()
        newton.linesearch.options["bound_enforcement"] = "scalar"
        newton.linesearch.options["iprint"] = -1
        newton.options["err_on_non_converge"] = False

        isolated_mass.linear_solver = om.DirectSolver(assemble_jac=True)



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import \
    distributed_engine_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class TransportInstrumentMass(om.ExplicitComponent):
    '''
    Calculates mass of instrument group for transports and GA aircraft.
    The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.

    ASSUMPTIONS: All engines have instrument mass that follows this equation
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.PLANFORM_AREA, 0.0)
        add_aviary_input(self, Aircraft.Instruments.MASS_SCALER, 1.0)

        add_aviary_output(self, Aircraft.Instruments.MASS, 0.0)

        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_crew = aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)
        num_wing_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES)
        num_fuse_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)
        num_wing_eng_fact = distributed_engine_count_factor(num_wing_eng)
        num_fuse_eng_fact = distributed_engine_count_factor(num_fuse_eng)

        fuse_area = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)
        mass_scaler = inputs[Aircraft.Instruments.MASS_SCALER]

        instrument_weight = (
            0.48 * fuse_area**0.57 * max_mach**0.5
            * (10.0 + 2.5 * num_crew + num_wing_eng_fact + 1.5 * num_fuse_eng_fact)
        )

        outputs[Aircraft.Instruments.MASS] = instrument_weight * \
            mass_scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_crew = aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)
        num_wing_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES)
        num_fuse_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)
        num_wing_eng_fact = distributed_engine_count_factor(num_wing_eng)
        num_fuse_eng_fact = distributed_engine_count_factor(num_fuse_eng)

        fuse_area = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)
        mass_scaler = inputs[Aircraft.Instruments.MASS_SCALER]

        fact = (10.0 + 2.5 * num_crew + num_wing_eng_fact + 1.5 * num_fuse_eng_fact)
        area_fact = fuse_area**0.57
        mach_fact = max_mach**0.5

        J[Aircraft.Instruments.MASS, Aircraft.Fuselage.PLANFORM_AREA] = (
            0.2736 * fuse_area**-0.43 * mach_fact * fact * mass_scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Instruments.MASS, Aircraft.Instruments.MASS_SCALER] = (
            0.48 * area_fact * mach_fact * fact / GRAV_ENGLISH_LBM
        )




import numpy as np
import openmdao.api as om
from openmdao.components.interp_util.interp import InterpND

from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class CompressibilityDrag(om.ExplicitComponent):
    """
    Computes compressibility drag coefficient.
    """

    def initialize(self):

        self.options.declare("num_nodes", default=6, types=int,
                             desc="Number of nodes along mission segment")

    def setup(self):
        nn = self.options["num_nodes"]

        # Simulation inputs
        self.add_input(
            Dynamic.Mission.MACH,
            shape=(nn),
            units='unitless',
            desc="Mach number")

        # Aero design inputs
        add_aviary_input(self, Mission.Design.MACH, 0.0)

        # Aircraft design inputs
        add_aviary_input(self, Aircraft.Design.BASE_AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, 0.0)
        add_aviary_input(self, Aircraft.Wing.SWEEP, 0.0)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.CROSS_SECTION, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.DIAMETER_TO_WING_SPAN, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.LENGTH_TO_DIAMETER, 0.0)

        # Outputs
        self.add_output('compress_drag_coeff', shape=(nn, ), units='unitless',
                        desc="Drag coefficient due to compressibility.")

    def setup_partials(self):
        nn = self.options["num_nodes"]

        row_col = np.arange(nn)
        self.declare_partials(of='compress_drag_coeff', wrt=[Dynamic.Mission.MACH],
                              rows=row_col, cols=row_col)

        wrt2 = [Aircraft.Wing.THICKNESS_TO_CHORD, Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SWEEP, Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
                Aircraft.Fuselage.CROSS_SECTION, Aircraft.Wing.AREA,
                Aircraft.Fuselage.LENGTH_TO_DIAMETER, Aircraft.Wing.TAPER_RATIO,
                Aircraft.Fuselage.DIAMETER_TO_WING_SPAN, Aircraft.Design.BASE_AREA,
                Mission.Design.MACH]

        self.declare_partials(of='compress_drag_coeff', wrt=wrt2)

    def compute(self, inputs, outputs):
        """
        Calculate compressibility drag.
        """

        del_mach = inputs[Dynamic.Mission.MACH] - inputs[Mission.Design.MACH]

        idx_super = np.where(del_mach > 0.05)
        idx_sub = np.where(del_mach <= 0.05)
        cdc_super = self._compute_supersonic(inputs, outputs, idx_super)
        cdc_sub = self._compute_subsonic(inputs, outputs, idx_sub)

        outputs['compress_drag_coeff'][idx_super] = cdc_super
        outputs['compress_drag_coeff'][idx_sub] = cdc_sub

    def _compute_supersonic(self, inputs, outputs, idx):
        """
        Calculate compressibility drag for supersonic speeds.
        """

        mach = inputs[Dynamic.Mission.MACH][idx]
        nn = len(mach)
        del_mach = mach - inputs[Mission.Design.MACH]
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        TC = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        max_camber_70 = inputs[Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN]
        sweep25 = inputs[Aircraft.Wing.SWEEP]
        wing_taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        fuse_area = inputs[Aircraft.Fuselage.CROSS_SECTION]
        base_area = inputs[Aircraft.Design.BASE_AREA]
        wing_area = inputs[Aircraft.Wing.AREA]
        fuselage_len_to_diam_ratio = inputs[Aircraft.Fuselage.LENGTH_TO_DIAMETER]
        diam_to_wing_span_ratio = inputs[Aircraft.Fuselage.DIAMETER_TO_WING_SPAN]

        ART = AR * np.tan(sweep25 / 57.2958) \
            + (1.0 - wing_taper_ratio) / (1.0 + wing_taper_ratio)
        x = np.empty((nn, 2), dtype=mach.dtype)
        x[:, 0] = del_mach
        x[:, 1] = ART
        CD3, self.dCD3 = PCARtable.interpolate(x, compute_derivative=True)

        # Negative drag sometimes occurs due to overshoot in the table interp.
        self.clamp_CD3 = np.where(CD3 <= 0)
        CD3[self.clamp_CD3] = 0.0
        self.dCD3[self.clamp_CD3] = 0.0

        compress_drag_coeff = CD3 * (TC ** (5.0 / 3.0) * (1.0 + 0.1 * max_camber_70))

        # Contribution of fuselage.
        if fuse_area > 0.0:
            SOS = 1.0 + base_area / fuse_area
            x[:, 0] = mach
            x[:, 1] = SOS
            CD4, self.dCD4 = BSUPtable.interpolate(x, compute_derivative=True)

            # Negative drag sometimes occurs due to overshoot in the table interp.
            self.clamp_CD4 = np.where(CD4 <= 0)
            CD4[self.clamp_CD4] = 0.0
            self.dCD4[self.clamp_CD4] = 0.0

            fuselage_compress_drag_coeff = CD4 * \
                (fuse_area / wing_area * (1.0 / fuselage_len_to_diam_ratio ** 2))

            compress_drag_coeff += fuselage_compress_drag_coeff

            # Wing fuselage interference.
            idx_mach = np.where(mach >= 1.0)
            if len(idx_mach[0]) > 0:

                x[:, 1] = diam_to_wing_span_ratio
                CD5, self.dCD5 = WFITable.interpolate(
                    x[idx_mach], compute_derivative=True)

                # TODO: is this some kind of override?
                if wing_taper_ratio == 1.0:
                    wing_taper_ratio = 0.5

                int_compress_drag_coeff = CD5 * (1.0 / (1.0 - wing_taper_ratio)
                                                 / np.cos(sweep25 / 57.2958))

                compress_drag_coeff[idx_mach] += int_compress_drag_coeff

            else:
                CD5 = np.zeros((nn), dtype=mach.dtype)
                self.dCD5 = np.zeros((nn, 2), dtype=mach.dtype)

        else:
            CD4 = np.zeros((nn), dtype=mach.dtype)
            self.dCD4 = np.zeros((nn, 2), dtype=mach.dtype)
            CD5 = np.zeros((nn), dtype=mach.dtype)
            self.dCD5 = np.zeros((nn, 2), dtype=mach.dtype)

        self.CD3 = CD3
        self.CD4 = CD4
        self.CD5 = CD5
        return compress_drag_coeff

    def _compute_subsonic(self, inputs, outputs, idx):
        """
        Calculate compressibility drag for subsonic speeds.
        """

        mach = inputs[Dynamic.Mission.MACH][idx]
        nn = len(mach)
        del_mach = mach - inputs[Mission.Design.MACH]
        TC = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        max_camber_70 = inputs[Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN]
        fuse_area = inputs[Aircraft.Fuselage.CROSS_SECTION]
        base_area = inputs[Aircraft.Design.BASE_AREA]
        wing_area = inputs[Aircraft.Wing.AREA]
        fuselage_len_to_diam_ratio = inputs[Aircraft.Fuselage.LENGTH_TO_DIAMETER]

        TOC = TC ** (2.0 / 3.0)
        x = np.empty((nn, 2), dtype=mach.dtype)
        x[:, 0] = del_mach
        x[:, 1] = TOC
        CD1, self.dCD1 = PCWtable.interpolate(x, compute_derivative=True)

        # Negative drag sometimes occurs due to overshoot in the table interp.
        self.clamp_CD1 = np.where(CD1 <= 0)
        CD1[self.clamp_CD1] = 0.0
        self.dCD1[self.clamp_CD1] = 0.0

        compress_drag_coeff = CD1 * (TC ** (5.0 / 3.0) * (1.0 + 0.1 * max_camber_70))

        # Contribution of fuselage.
        if fuse_area > 0.0:
            SOS = 1.0 + base_area / fuse_area
            x[:, 0] = mach
            x[:, 1] = SOS
            CD2, self.dCD2 = BSUBtable.interpolate(x, compute_derivative=True)

            # Negative drag sometimes occurs due to overshoot in the table interp.
            self.clamp_CD2 = np.where(CD2 <= 0)
            CD2[self.clamp_CD2] = 0.0
            self.dCD2[self.clamp_CD2] = 0.0

            fuselage_compress_drag_coeff = CD2 * \
                (fuse_area / wing_area * (1.0 / fuselage_len_to_diam_ratio ** 2))

            compress_drag_coeff += fuselage_compress_drag_coeff

        else:
            CD2 = np.zeros((nn), dtype=mach.dtype)
            self.dCD2 = np.zeros((nn, 2), dtype=mach.dtype)

        self.CD1 = CD1
        self.CD2 = CD2
        return compress_drag_coeff

    def compute_partials(self, inputs, partials):
        """
        Calculate partials of compressibility drag.

        :param inputs: _description_
        :type inputs: _type_
        :param partials: _description_
        :type partials: _type_
        """

        del_mach = inputs[Dynamic.Mission.MACH] - inputs[Mission.Design.MACH]

        idx_super = np.where(del_mach > 0.05)
        idx_sub = np.where(del_mach <= 0.05)
        self._compute_partials_supersonic(inputs, partials, idx_super)
        self._compute_partials_subsonic(inputs, partials, idx_sub)

    def _compute_partials_supersonic(self, inputs, partials, idx):

        mach = inputs[Dynamic.Mission.MACH][idx]
        nn = len(mach)
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        TC = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        max_camber_70 = inputs[Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN]
        sweep25 = inputs[Aircraft.Wing.SWEEP]
        wing_taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        fuse_area = inputs[Aircraft.Fuselage.CROSS_SECTION]
        base_area = inputs[Aircraft.Design.BASE_AREA]
        wing_area = inputs[Aircraft.Wing.AREA]
        fuselage_len_to_diam_ratio = inputs[Aircraft.Fuselage.LENGTH_TO_DIAMETER]

        CD3 = self.CD3
        CD4 = self.CD4
        CD5 = self.CD5

        dCd3_dCD3 = TC ** (5.0 / 3.0) * (1.0 + max_camber_70 / 10.0)
        dCd4_dCD4 = (fuse_area / wing_area) * (1.0 / fuselage_len_to_diam_ratio ** 2)
        dCd5_dCD5 = 1.0 / (1.0 - wing_taper_ratio) / np.cos(sweep25 / 57.2958)

        ddel_mach_ddesign_Mach = -1.0
        ddel_mach_dMach = 1.0

        # Derivatives from table interpolation
        dCD3_ddel_mach = self.dCD3[:, 0]
        dCD3_dART = self.dCD3[:, 1]
        dCD4_dMach = self.dCD4[:, 0]
        dCD4_dSOS = self.dCD4[:, 1]
        dCD5_dMach = self.dCD5[:, 0]
        dCD5_ddiam_to_wing_span_ratio = self.dCD5[:, 1]

        # wrt Mach
        dCd3_dMach = dCd3_dCD3 * dCD3_ddel_mach * ddel_mach_dMach

        # wrt design_Mach
        dCd_ddesign_Mach = dCd3_dCD3 * dCD3_ddel_mach * ddel_mach_ddesign_Mach

        # wrt TC
        dCd3_dTC = (5.0 / 3.0) * CD3 * TC**(2.0 / 3.0) * (1.0 + 0.1 * max_camber_70)
        dCd_dTC = dCd3_dTC

        # wrt max_camber_70
        dCd3_dmax_camber_70 = 0.1 * CD3 * TC**(5.0 / 3.0)
        dCd_dmax_camber_70 = dCd3_dmax_camber_70

        # wrt AR
        dART_dAR = np.tan(sweep25 / 57.2958)
        dCd_dAR = dCd3_dCD3 * dCD3_dART * dART_dAR

        dART_dwing_taper_ratio = -(1 - wing_taper_ratio) / (wing_taper_ratio + 1)**2 \
            - 1.0 / (wing_taper_ratio + 1)

        dART_dsweep25 = AR * (np.tan(sweep25 / 57.2958)**2 + 1) / 57.2958
        dCd3_dsweep25 = dCd3_dCD3 * dCD3_dART * dART_dsweep25

        dCd_ddiam_to_wing_span_ratio = 0

        if fuse_area > 0.0:

            # wrt Mach
            dCd4_dMach = dCd4_dCD4 * dCD4_dMach
            dCd_dMach = dCd3_dMach + dCd4_dMach

            # wrt fuse_area
            dSOS_dfuse_area = -base_area / fuse_area**2
            dCd4_dfuse_area = CD4 / (wing_area * fuselage_len_to_diam_ratio**2)
            dCd_dfuse_area = dCd4_dfuse_area + dCd4_dCD4 * dCD4_dSOS * dSOS_dfuse_area

            # wrt base_area
            dSOS_dbase_area = 1.0 / fuse_area
            dCd_dbase_area = dCd4_dCD4 * dCD4_dSOS * dSOS_dbase_area

            # wrt wing_area
            dCd_dwing_area = -CD4 * fuse_area / \
                (wing_area * fuselage_len_to_diam_ratio)**2

            # wrt fuselage_len_to_diam_ratio
            dCd_dfuselage_len_to_diam_ratio = -2.0 * CD4 * fuse_area \
                / (wing_area * fuselage_len_to_diam_ratio**3)

            # wrt diam_to_wing_span_ratio
            dCd_ddiam_to_wing_span_ratio = np.zeros(dCd_dwing_area.shape,
                                                    dtype=dCd_dwing_area.dtype)

            # wrt wing_taper_ratio
            dCd_dwing_taper_ratio = dCd3_dCD3 * dCD3_dART * dART_dwing_taper_ratio

            # wrt SW25
            dCd_dsweep25 = dCd3_dsweep25

            # Wing fuselage interference.
            idx_mach = np.where(mach >= 1.0)
            if len(idx_mach[0]) > 0:

                dCd_dMach[idx_mach] += dCd5_dCD5 * dCD5_dMach

                dCd5_dwing_taper_ratio = CD5 / \
                    ((1.0 - wing_taper_ratio)**2 * np.cos(sweep25 / 57.2958))
                dCd_dwing_taper_ratio[idx_mach] += dCd5_dwing_taper_ratio

                # wrt diam_to_wing_span_ratio
                dCd_ddiam_to_wing_span_ratio[idx_mach] = dCd5_dCD5 * \
                    dCD5_ddiam_to_wing_span_ratio

                dCd5_dsweep25 = CD5 * np.sin(sweep25 / 57.2958) / \
                    (57.2958 * (1.0 - wing_taper_ratio) * np.cos(sweep25 / 57.2958)**2)
                dCd_dsweep25[idx_mach] += dCd5_dsweep25

        else:

            dCd_dMach = dCd3_dMach
            dCd_dfuse_area = 0.0
            dCd_dbase_area = 0.0
            dCd_dwing_area = 0.0
            dCd_dfuselage_len_to_diam_ratio = 0.0
            dCd_dwing_taper_ratio = dCd3_dCD3 * dCD3_dART * dART_dwing_taper_ratio
            dCd_dsweep25 = dCd3_dsweep25

        partials["compress_drag_coeff", Dynamic.Mission.MACH][idx] = dCd_dMach
        partials["compress_drag_coeff", Mission.Design.MACH][idx, 0] = dCd_ddesign_Mach
        partials["compress_drag_coeff", Aircraft.Wing.THICKNESS_TO_CHORD][idx, 0] = dCd_dTC
        partials["compress_drag_coeff",
                 Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN][idx, 0] = dCd_dmax_camber_70
        partials["compress_drag_coeff",
                 Aircraft.Fuselage.CROSS_SECTION][idx, 0] = dCd_dfuse_area
        partials["compress_drag_coeff",
                 Aircraft.Design.BASE_AREA][idx,
                                            0] = dCd_dbase_area
        partials["compress_drag_coeff", Aircraft.Wing.AREA][idx, 0] = dCd_dwing_area
        partials["compress_drag_coeff",
                 Aircraft.Fuselage.LENGTH_TO_DIAMETER][idx,
                                                       0] = dCd_dfuselage_len_to_diam_ratio
        partials["compress_drag_coeff",
                 Aircraft.Wing.TAPER_RATIO][idx, 0] = dCd_dwing_taper_ratio
        partials["compress_drag_coeff", Aircraft.Wing.SWEEP][idx, 0] = dCd_dsweep25
        partials["compress_drag_coeff", Aircraft.Wing.ASPECT_RATIO][idx, 0] = dCd_dAR
        partials["compress_drag_coeff",
                 Aircraft.Fuselage.DIAMETER_TO_WING_SPAN][idx,
                                                          0] = dCd_ddiam_to_wing_span_ratio

    def _compute_partials_subsonic(self, inputs, partials, idx):

        mach = inputs[Dynamic.Mission.MACH][idx]
        nn = len(mach)
        TC = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        max_camber_70 = inputs[Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN]
        fuse_area = inputs[Aircraft.Fuselage.CROSS_SECTION]
        base_area = inputs[Aircraft.Design.BASE_AREA]
        wing_area = inputs[Aircraft.Wing.AREA]
        fuselage_len_to_diam_ratio = inputs[Aircraft.Fuselage.LENGTH_TO_DIAMETER]
        CD1 = self.CD1
        CD2 = self.CD2

        dCd1_dCD1 = TC ** (5.0 / 3.0) * (1.0 + 0.1 * max_camber_70)
        dCd2_dCD2 = fuse_area / wing_area * (1.0 / fuselage_len_to_diam_ratio ** 2)
        dTOC_dTC = (2.0 / 3.0) * TC ** (-1.0 / 3.0)

        ddel_mach_dMach = 1.0

        # Derivatives from table interpolation
        dCD1_ddel_mach = self.dCD1[:, 0]
        dCD1_dTOC = self.dCD1[:, 1]
        dCD2_dMach = self.dCD2[:, 0]
        dCD2_dSOS = self.dCD2[:, 1]

        # wrt Mach
        dCd1_dMach = dCd1_dCD1 * dCD1_ddel_mach * ddel_mach_dMach
        dCd2_dMach = dCd2_dCD2 * dCD2_dMach
        dCd_dMach = dCd1_dMach + dCd2_dMach

        # wrt design_Mach
        dCd_ddesign_Mach = -dCd1_dCD1 * dCD1_ddel_mach

        # wrt TC
        dCd1_dTC = (5.0 / 3.0) * CD1 * (1.0 + 0.1 * max_camber_70) * TC**(2.0 / 3.0)
        dCD1_dTC = dCD1_dTOC * dTOC_dTC
        dCd_dTC = dCd1_dTC + dCd1_dCD1 * dCD1_dTC

        # wrt max_camber_70
        dCd_dmax_camber_70 = CD1 * (1.0 / 10.0) * TC**(5.0 / 3.0)

        partials["compress_drag_coeff", Dynamic.Mission.MACH][idx] = dCd_dMach
        partials["compress_drag_coeff", Mission.Design.MACH][idx, 0] = dCd_ddesign_Mach
        partials["compress_drag_coeff", Aircraft.Wing.THICKNESS_TO_CHORD][idx, 0] = dCd_dTC
        partials["compress_drag_coeff",
                 Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN][idx, 0] = dCd_dmax_camber_70

        # Contribution of fuselage.
        if fuse_area > 0.0:

            dCd2_dfuse_area = CD2 / (wing_area * fuselage_len_to_diam_ratio**2)
            dSOS_dfuse_area = -base_area / fuse_area**2
            dCd_dfuse_area = dCd2_dfuse_area + dCd2_dCD2 * dCD2_dSOS * dSOS_dfuse_area

            # wrt base_area
            dSOS_dbase_area = 1.0 / fuse_area
            dCd_dbase_area = dCd2_dCD2 * dCD2_dSOS * dSOS_dbase_area

            # wrt wing_area
            dCd_dwing_area = -CD2 * fuse_area / \
                (wing_area * fuselage_len_to_diam_ratio)**2

            # wrt fuselage_len_to_diam_ratio
            dCd_dfuselage_len_to_diam_ratio = -2.0 * CD2 * fuse_area \
                / (wing_area * fuselage_len_to_diam_ratio**3)

            partials["compress_drag_coeff",
                     Aircraft.Fuselage.CROSS_SECTION][idx, 0] = dCd_dfuse_area
            partials["compress_drag_coeff",
                     Aircraft.Design.BASE_AREA][idx, 0] = dCd_dbase_area
            partials["compress_drag_coeff", Aircraft.Wing.AREA][idx, 0] = dCd_dwing_area
            partials["compress_drag_coeff",
                     Aircraft.Fuselage.LENGTH_TO_DIAMETER][idx,
                                                           0] = dCd_dfuselage_len_to_diam_ratio

        else:
            partials["compress_drag_coeff",
                     Aircraft.Fuselage.CROSS_SECTION][idx, 0] = 0.0
            partials["compress_drag_coeff", Aircraft.Design.BASE_AREA][idx, 0] = 0.0
            partials["compress_drag_coeff", Aircraft.Wing.AREA][idx, 0] = 0.0
            partials["compress_drag_coeff",
                     Aircraft.Fuselage.LENGTH_TO_DIAMETER][idx, 0] = 0.0

        partials["compress_drag_coeff", Aircraft.Wing.TAPER_RATIO][idx, 0] = 0.0
        partials["compress_drag_coeff", Aircraft.Wing.SWEEP][idx, 0] = 0.0
        partials["compress_drag_coeff", Aircraft.Wing.ASPECT_RATIO][idx, 0] = 0.0
        partials["compress_drag_coeff",
                 Aircraft.Fuselage.DIAMETER_TO_WING_SPAN][idx, 0] = 0.0


# Tables
PCW = np.array([[13007.0, .100, .120, .140, .160, .180, .220, .300],
                [-.800, .00, .00, .00, .00, .00, .00, .00],
                [-.200, .0600, .040, .020, .0200, .0100, .0080, .0020],
                [-.160, .0720, .050, .030, .0260, .0170, .0160, .0060],
                [-.120, .1000, .060, .040, .0380, .0250, .0240, .0120],
                [-.080, .1250, .080, .050, .0490, .0350, .0330, .0190],
                [-.040, .1600, .120, .080, .0680, .0540, .0470, .0300],
                [-.020, .2000, .160, .120, .1100, .0700, .0590, .0390],
                [0.000, .2800, .220, .160, .1200, .0930, .0770, .0520],
                [.010, .3400, .270, .200, .1520, .1180, .0930, .0610],
                [.020, .4400, .330, .240, .1970, .1530, .1170, .0730],
                [.030, .6400, .450, .310, .2550, .2030, .1480, .0870],
                [.040, 1.1000, .660, .410, .3250, .2700, .1870, .1030],
                [.050, 1.9000, 1.020, .560, .4000, .3500, .2350, .1270]])

BSUB = np.array([[17004.0, 1.00, 1.20, 1.40, 1.50],
                 [.2000, 0.00, 0.00, 0.00, 0.00],
                 [.5000, 0.00, 0.00, 0.00, 0.00],
                 [.7000, 0.00, 0.00, 0.00, 0.00],
                 [.7800, 0.00, 0.00, 0.00, 0.00],
                 [.8200, 0.00, 0.00, 0.150, 0.210],
                 [.8400, 0.00, 0.150, 0.200, 0.350],
                 [.8600, 0.090, 0.220, 0.400, 0.520],
                 [.8800, 0.200, 0.380, 0.610, 0.780],
                 [.9000, 0.380, 0.580, 0.910, 1.100],
                 [.9100, 0.530, 0.750, 1.100, 1.330],
                 [.9200, 0.730, 0.950, 1.300, 1.600],
                 [.9300, 0.950, 1.200, 1.650, 1.930],
                 [.9400, 1.300, 1.550, 2.050, 2.490],
                 [.9500, 1.750, 2.200, 2.900, 3.650],
                 [.9600, 2.450, 3.250, 4.500, 6.400],
                 [.9650, 3.000, 4.220, 6.300, 8.450],
                 [.9700, 3.900, 5.600, 9.500, 11.500]])

PCAR = np.array([[16009.0, 1.00, 1.50, 2.00, 2.50, 3.00, 3.50, 4.00, 5.00, 6.00],
                 [.050, 2.400, 1.700, 1.170, .850, .730, .670, .600, .540, .520],
                 [.070, 3.100, 2.250, 1.580, 1.100, .890, .770, .700, .620, .600],
                 [.090, 3.550, 2.610, 1.880, 1.240, .990, .870, .750, .670, .650],
                 [.110, 3.850, 2.880, 2.030, 1.330, 1.070, .920, .800, .710, .680],
                 [.130, 3.970, 3.050, 2.140, 1.410, 1.120, .960, .840, .740, .710],
                 [.150, 4.000, 3.100, 2.170, 1.480, 1.160, .990, .860, .750, .720],
                 [.200, 3.900, 3.000, 2.200, 1.550, 1.200, 1.000, .860, .740, .700],
                 [.250, 3.680, 2.850, 2.160, 1.570, 1.200, 1.000, .830, .700, .650],
                 [.300, 3.430, 2.700, 2.100, 1.550, 1.170, .920, .770, .630, .580],
                 [.400, 3.030, 2.450, 1.900, 1.470, 1.100, .880, .730, .590, .530],
                 [.500, 2.750, 2.220, 1.710, 1.370, 1.020, .840, .730, .570, .520],
                 [.600, 2.490, 2.000, 1.550, 1.260, .970, .810, .740, .560, .510],
                 [.700, 2.250, 1.800, 1.410, 1.170, .910, .790, .710, .550, .510],
                 [.800, 1.990, 1.620, 1.300, 1.100, .880, .750, .700, .550, .500],
                 [.900, 1.800, 1.500, 1.200, 1.000, .840, .700, .660, .540, .500],
                 [1.000, 1.650, 1.400, 1.100, .950, .800, .700, .660, .540, .500]])

BSUP = np.array([[14006.0, 1.00, 1.10, 1.20, 1.30, 1.40, 1.50],
                 [1.000, 24.50, 20.00, 16.20, 13.40, 11.10, 9.50],
                 [1.050, 30.70, 23.60, 20.00, 16.00, 12.90, 10.50],
                 [1.100, 33.00, 26.20, 21.50, 17.40, 14.00, 11.10],
                 [1.150, 34.30, 27.30, 22.30, 18.20, 14.80, 11.60],
                 [1.200, 34.70, 27.70, 22.50, 18.50, 15.00, 11.90],
                 [1.250, 34.50, 27.50, 22.40, 18.20, 14.90, 11.90],
                 [1.300, 33.80, 27.00, 22.00, 17.60, 14.50, 11.70],
                 [1.350, 32.90, 26.40, 21.70, 17.30, 14.20, 11.40],
                 [1.400, 32.40, 25.90, 21.40, 17.20, 14.10, 11.00],
                 [1.500, 32.00, 25.60, 21.10, 17.00, 14.10, 10.90],
                 [1.600, 32.00, 25.60, 21.00, 17.00, 14.10, 10.90],
                 [1.800, 32.00, 25.60, 21.00, 17.00, 14.20, 11.40],
                 [2.000, 32.00, 25.60, 21.00, 17.10, 14.40, 11.80],
                 [2.200, 32.00, 25.60, 21.00, 17.30, 14.60, 12.00]])

# called BINT in FLOPS.
WFI = np.array([[13010.0, .10, .120, .140, .150, .160, .170, .180, .190, .200, .220],
                [1.000, 0.0, 0.0, 0.0, 0.0, 0.0,
                    0.0, 0.0, 0.0, 0.0, 0.0],
                [1.050, 0.0, 0.0, .00040, -.00030, -.00080, -.00110, -.00100, -.00040, .00030, .00180],
                [1.100, 0.0, 0.0, .00060, -.00060, -.00140, -.00180, -.00140, -.00060, .00040, .00260],
                [1.150, 0.0, 0.0, .00030, -.00080, -.00170, -.00200, -.00150, -.00060, .00040, .00240],
                [1.200, 0.0, 0.0, .00020, -.00080, -.00170, -.00180, -.00140, -.00060, .00030, .00200],
                [1.300, 0.0, 0.0, .00020, -.00060, -.00100, -.00100, -.00080, -.00050, .00010, .00120],
                [1.400, 0.0, 0.0, .00010, -.00030, -.00030, -.00030, -.00020, -.00010, .00030, .00090],
                [1.500, 0.0, 0.0, .00010, .00000, .00030, .00030, .00040, .00040, .00050, .00070],
                [1.600, 0.0, 0.0, .00000, .00040, .00050, .00090, .00090, .00080, .00070, .00050],
                [1.700, 0.0, 0.0, .00000, .00050, .00070, .00120, .00110, .00100, .00080, .00050],
                [1.800, 0.0, 0.0, .00000, .00060, .00090, .00120, .00110, .00100, .00080, .00050],
                [1.900, 0.0, 0.0, .00000, .00060, .00090, .00100, .00100, .00090, .00080, .00050],
                [2.000, 0.0, 0.0, .00000, .00050, .00090, .00110, .00100, .00090, .00070, .00050]])

PCWtable = InterpND(method='lagrange2', points=(
    PCW[1:, 0], PCW[0, 1:]), values=PCW[1:, 1:], extrapolate=True)
BSUBtable = InterpND(method='lagrange2', points=(
    BSUB[1:, 0], BSUB[0, 1:]), values=BSUB[1:, 1:], extrapolate=True)
PCARtable = InterpND(method='lagrange2', points=(
    PCAR[1:, 0], PCAR[0, 1:]), values=PCAR[1:, 1:], extrapolate=True)
BSUPtable = InterpND(method='lagrange2', points=(
    BSUP[1:, 0], BSUP[0, 1:]), values=BSUP[1:, 1:], extrapolate=True)
WFITable = InterpND(method='lagrange2', points=(
    WFI[1:, 0], WFI[0, 1:]), values=WFI[1:, 1:], extrapolate=True)



import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.mission.gasp_based.phases.breguet import RangeComp
from aviary.subsystems.mass.mass_to_weight import MassToWeight
from aviary.subsystems.propulsion.propulsion_builder import PropulsionBuilderBase
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Dynamic
from aviary.mission.ode.specific_energy_rate import SpecificEnergyRate
from aviary.mission.ode.altitude_rate import AltitudeRate


class BreguetCruiseODESolution(BaseODE):
    def setup(self):
        nn = self.options["num_nodes"]
        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['core_subsystems']

        # TODO: paramport
        self.add_subsystem("params", ParamPort(), promotes=["*"])

        self.add_subsystem(
            "USatm",
            USatm1976Comp(
                num_nodes=nn),
            promotes_inputs=[
                ("h",
                 Dynamic.Mission.ALTITUDE)],
            promotes_outputs=[
                "rho",
                ("sos",
                 Dynamic.Mission.SPEED_OF_SOUND),
                ("temp",
                 Dynamic.Mission.TEMPERATURE),
                ("pres",
                 Dynamic.Mission.STATIC_PRESSURE),
                "viscosity"],
        )

        self.add_subsystem(
            "fc",
            FlightConditions(
                num_nodes=nn,
                input_speed_type=SpeedType.MACH),
            promotes_inputs=[
                "rho",
                Dynamic.Mission.SPEED_OF_SOUND,
                Dynamic.Mission.MACH],
            promotes_outputs=[
                Dynamic.Mission.DYNAMIC_PRESSURE,
                "EAS",
                ("TAS", Dynamic.Mission.VELOCITY)],
        )

        self.add_subsystem(
            "calc_weight",
            MassToWeight(num_nodes=nn),
            promotes_inputs=["mass"],
            promotes_outputs=["weight"]
        )

        prop_group = om.Group()

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options,
                  'method': 'cruise', 'output_alpha': True}
        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                if isinstance(subsystem, PropulsionBuilderBase):
                    prop_group.add_subsystem(subsystem.name,
                                             system,
                                             promotes_inputs=subsystem.mission_inputs(
                                                 **kwargs),
                                             promotes_outputs=subsystem.mission_outputs(**kwargs))
                else:
                    self.add_subsystem(subsystem.name,
                                       system,
                                       promotes_inputs=subsystem.mission_inputs(
                                           **kwargs),
                                       promotes_outputs=subsystem.mission_outputs(**kwargs))

        bal = om.BalanceComp(
            name=Dynamic.Mission.THROTTLE,
            val=np.ones(nn),
            upper=1.0,
            lower=0.0,
            units="unitless",
            lhs_name=Dynamic.Mission.THRUST_TOTAL,
            rhs_name=Dynamic.Mission.DRAG,
            eq_units="lbf",
        )

        prop_group.add_subsystem(
            "thrust_balance",
            subsys=bal,
            promotes_inputs=['*'],
            promotes_outputs=['*'])

        prop_group.linear_solver = om.DirectSolver()

        prop_group.nonlinear_solver = om.NewtonSolver(
            solve_subsystems=True,
            maxiter=20,
            rtol=1e-12,
            atol=1e-12,
            err_on_non_converge=False,
        )
        prop_group.nonlinear_solver.linesearch = om.BoundsEnforceLS()

        prop_group.nonlinear_solver.options["iprint"] = 2
        prop_group.linear_solver.options["iprint"] = 2

        self.add_subsystem(
            'prop_group',
            subsys=prop_group,
            promotes_inputs=['*'],
            promotes_outputs=['*'])

        #
        # collect initial/final outputs
        #
        self.add_subsystem(
            "eom",
            RangeComp(num_nodes=nn),
            promotes_inputs=[
                ("cruise_distance_initial", "initial_distance"),
                ("cruise_time_initial", "initial_time"),
                "mass",
                Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
                ("TAS_cruise", Dynamic.Mission.VELOCITY),
            ],
            promotes_outputs=[("cruise_range", Dynamic.Mission.DISTANCE),
                              ("cruise_time", "time")],
        )

        self.add_subsystem(
            name='SPECIFIC_ENERGY_RATE_EXCESS',
            subsys=SpecificEnergyRate(num_nodes=nn),
            promotes_inputs=[Dynamic.Mission.VELOCITY, Dynamic.Mission.MASS,
                             (Dynamic.Mission.THRUST_TOTAL, Dynamic.Mission.THRUST_MAX_TOTAL),
                             Dynamic.Mission.DRAG],
            promotes_outputs=[(Dynamic.Mission.SPECIFIC_ENERGY_RATE,
                               Dynamic.Mission.SPECIFIC_ENERGY_RATE_EXCESS)]
        )

        self.add_subsystem(
            name='ALTITUDE_RATE_MAX',
            subsys=AltitudeRate(num_nodes=nn),
            promotes_inputs=[
                (Dynamic.Mission.SPECIFIC_ENERGY_RATE,
                 Dynamic.Mission.SPECIFIC_ENERGY_RATE_EXCESS),
                Dynamic.Mission.VELOCITY_RATE,
                Dynamic.Mission.VELOCITY],
            promotes_outputs=[
                (Dynamic.Mission.ALTITUDE_RATE,
                 Dynamic.Mission.ALTITUDE_RATE_MAX)])

        ParamPort.set_default_vals(self)
        self.set_input_defaults(
            Dynamic.Mission.ALTITUDE,
            val=37500 * np.ones(nn),
            units="ft")
        self.set_input_defaults("mass", val=np.linspace(
            171481, 171581 - 10000, nn), units="lbm")



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.fuel_capacity import (
    AuxFuelCapacity, FuelCapacityGroup, FuselageFuelCapacity,
    TotalFuelCapacity, WingFuelCapacity)
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (do_validation_test,
                                                      flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.functions import override_aviary_vars
from aviary.variable_info.variables import Aircraft


class FuelCapacityGroupTest(unittest.TestCase):

    @parameterized.expand(get_flops_case_names(only=['N3CC']),
                          name_func=print_case)
    def test_case(self, case_name):

        class StaticsHarness(om.Group):

            def initialize(self):
                self.options.declare(
                    'aviary_options', types=AviaryValues,
                    desc='Aircraft options dictionary')

            def setup(self):
                aviary_options = self.options['aviary_options']

                self.add_subsystem(
                    'fuel_capy_group', FuelCapacityGroup(aviary_options=aviary_options),
                    promotes_inputs=['*'], promotes_outputs=['*'])

            def configure(self):
                aviary_options = self.options['aviary_options']

                # Overrides
                override_aviary_vars(self, aviary_options)

        prob = om.Problem()
        prob.model.add_subsystem(
            "statics",
            StaticsHarness(aviary_options=get_flops_inputs(case_name)),
            promotes_outputs=['*'],
            promotes_inputs=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
                        Aircraft.Fuel.CAPACITY_FACTOR,
                        Aircraft.Fuel.DENSITY_RATIO,
                        Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY,
                        Aircraft.Wing.AREA,
                        Aircraft.Wing.SPAN,
                        Aircraft.Wing.TAPER_RATIO,
                        Aircraft.Wing.THICKNESS_TO_CHORD],
            output_keys=Aircraft.Fuel.TOTAL_CAPACITY,
            atol=1e-10,
            rtol=1e-10)


wing_capacity_data = {}
wing_capacity_data['1'] = AviaryValues({
    Aircraft.Fuel.DENSITY_RATIO: (1.2, 'unitless'),
    Aircraft.Fuel.WING_REF_CAPACITY: (30.0, 'lbm'),
    Aircraft.Fuel.WING_REF_CAPACITY_AREA: (200.0, 'unitless'),
    Aircraft.Fuel.WING_REF_CAPACITY_TERM_B: (1.3, 'unitless'),
    Aircraft.Fuel.CAPACITY_FACTOR: (1.0, 'unitless'),
    Aircraft.Wing.AREA: (150.0, 'ft**2'),
    Aircraft.Wing.SPAN: (17.0, 'ft'),
    Aircraft.Wing.TAPER_RATIO: (1.5, 'unitless'),
    Aircraft.Wing.THICKNESS_TO_CHORD: (0.33, 'unitless'),
    Aircraft.Fuel.WING_REF_CAPACITY_TERM_A: (-100.0, 'unitless')
})
wing_capacity_data['2'] = AviaryValues({
    Aircraft.Fuel.DENSITY_RATIO: (1.2, 'unitless'),
    Aircraft.Fuel.WING_REF_CAPACITY: (30.0, 'lbm'),
    Aircraft.Fuel.WING_REF_CAPACITY_AREA: (200.0, 'unitless'),
    Aircraft.Fuel.WING_REF_CAPACITY_TERM_B: (1.3, 'unitless'),
    Aircraft.Fuel.CAPACITY_FACTOR: (1.0, 'unitless'),
    Aircraft.Wing.AREA: (150.0, 'ft**2'),
    Aircraft.Wing.SPAN: (17.0, 'ft'),
    Aircraft.Wing.TAPER_RATIO: (1.5, 'unitless'),
    Aircraft.Wing.THICKNESS_TO_CHORD: (0.33, 'unitless'),
    Aircraft.Fuel.WING_REF_CAPACITY_TERM_A: (1.2, 'unitless')
})

wing_capacity_cases = [key for key in wing_capacity_data]


class WingFuelCapacityTest(unittest.TestCase):

    @parameterized.expand(wing_capacity_cases,
                          name_func=print_case)
    def test_derivs(self, case_name):
        validation_data = wing_capacity_data[case_name]
        prob = om.Problem()

        prob.model.add_subsystem(
            'comp',
            WingFuelCapacity(),
            promotes=['*']
        )

        prob.setup(force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=validation_data,
                           output_validation_data=validation_data,
                           input_keys=[
                               Aircraft.Fuel.DENSITY_RATIO,
                               Aircraft.Fuel.WING_REF_CAPACITY,
                               Aircraft.Fuel.WING_REF_CAPACITY_AREA,
                               Aircraft.Fuel.WING_REF_CAPACITY_TERM_B,
                               Aircraft.Fuel.CAPACITY_FACTOR,
                               Aircraft.Wing.AREA,
                               Aircraft.Wing.SPAN,
                               Aircraft.Wing.TAPER_RATIO,
                               Aircraft.Wing.THICKNESS_TO_CHORD,
                               Aircraft.Fuel.WING_REF_CAPACITY_TERM_A
                           ],
                           output_keys=Aircraft.Fuel.WING_FUEL_CAPACITY,
                           atol=1e-10,
                           # TODO: No wing fuel capacity validation data, check only partials
                           check_values=False)


fuse_capacity_data = {}
fuse_capacity_data['1'] = AviaryValues({
    Aircraft.Fuel.TOTAL_CAPACITY: (100.0, 'lbm'),
    Aircraft.Fuel.WING_FUEL_CAPACITY: (73.0, 'lbm'),
    Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY: (27.0, 'lbm')
})

fuse_capacity_cases = [key for key in fuse_capacity_data]


class FuselageFuelCapacityTest(unittest.TestCase):

    @parameterized.expand(fuse_capacity_cases,
                          name_func=print_case)
    def test_basic(self, case_name):
        validation_data = fuse_capacity_data[case_name]
        prob = om.Problem()

        prob.model.add_subsystem(
            'fuel',
            FuselageFuelCapacity(),
            promotes=['*']
        )

        prob.setup(force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=validation_data,
                           output_validation_data=validation_data,
                           input_keys=[
                               Aircraft.Fuel.TOTAL_CAPACITY,
                               Aircraft.Fuel.WING_FUEL_CAPACITY
                           ],
                           output_keys=Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY,
                           tol=1.0e-10,
                           atol=1e-10)


aux_capacity_data = {}
aux_capacity_data['1'] = AviaryValues({
    Aircraft.Fuel.TOTAL_CAPACITY: (100.0, 'lbm'),
    Aircraft.Fuel.WING_FUEL_CAPACITY: (25.0, 'lbm'),
    Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY: (33.0, 'lbm'),
    Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY: (42.0, 'lbm')
})

aux_capacity_cases = [key for key in aux_capacity_data]


class AuxFuelCapacityTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(aux_capacity_cases,
                          name_func=print_case)
    def test_basic(self, case_name):
        validation_data = aux_capacity_data[case_name]
        prob = self.prob

        prob.model.add_subsystem(
            'fuel',
            AuxFuelCapacity(),
            promotes=['*']
        )

        prob.setup(force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=validation_data,
                           output_validation_data=validation_data,
                           input_keys=[
                               Aircraft.Fuel.TOTAL_CAPACITY,
                               Aircraft.Fuel.WING_FUEL_CAPACITY,
                               Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY
                           ],
                           output_keys=Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
                           tol=1.0e-10,
                           atol=1e-10)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


total_capacity_data = {}
total_capacity_data['1'] = AviaryValues({
    Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY: (100.0, 'lbm'),
    Aircraft.Fuel.WING_FUEL_CAPACITY: (25.0, 'lbm'),
    Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY: (33.0, 'lbm'),
    Aircraft.Fuel.TOTAL_CAPACITY: (158.0, 'lbm')
})

total_capacity_cases = [key for key in total_capacity_data]


class TotalFuelCapacityTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(total_capacity_cases,
                          name_func=print_case)
    def test_basic(self, case_name):
        validation_data = total_capacity_data[case_name]
        prob = self.prob

        prob.model.add_subsystem(
            'fuel',
            TotalFuelCapacity(),
            promotes=['*']
        )

        prob.setup(force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=validation_data,
                           output_validation_data=validation_data,
                           input_keys=[
                               Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
                               Aircraft.Fuel.WING_FUEL_CAPACITY,
                               Aircraft.Fuel.FUSELAGE_FUEL_CAPACITY
                           ],
                           output_keys=Aircraft.Fuel.TOTAL_CAPACITY,
                           tol=1.0e-10,
                           atol=1e-10)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



"""
Fortran_to_Aviary.py is used to read in Fortran based vehicle decks and convert them to Aviary decks.

FLOPS, GASP, or Aviary names can be used for variables (Ex WG or Mission:Design:GROSS_MASS)
When specifying variables from FORTRAN, they should be in the appropriate NAMELIST.
Aviary variable names should be specified outside any NAMELISTS.
Names are not case-sensitive.
Units can be specified using any of the openMDAO valid units.
Comments can be added using !
Lists can be entered by separating values with commas.
Individual list elements can be specified by adding an index after the variable name.
(NOTE: 1 indexing is used inside NAMELISTS, while 0 indexing is used outside NAMELISTS)

Example inputs:
aircraft:fuselage:pressure_differential = .5, atm !DELP in GASP, but using atmospheres instead of psi
ARNGE(1) = 3600 !target range in nautical miles
pyc_phases = taxi, groundroll, rotation, landing
debug_mode = True
"""

import csv
import re
from enum import Enum
from pathlib import Path

from openmdao.utils.units import valid_units

from aviary.utils.functions import convert_strings_to_data
from aviary.utils.named_values import NamedValues, get_items
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.variables import Aircraft, Mission
from aviary.variable_info.enums import LegacyCode
from aviary.utils.functions import get_path
from aviary.utils.legacy_code_data.deprecated_vars import flops_deprecated_vars, gasp_deprecated_vars


FLOPS = LegacyCode.FLOPS
GASP = LegacyCode.GASP


def create_aviary_deck(fortran_deck: str, legacy_code=None, defaults_deck=None,
                       out_file=None, force=False):
    '''
    Create an Aviary CSV file from a Fortran input deck
    Required input is the filepath to the input deck and legacy code. Optionally, a
    deck of default values can be specified, this is useful if an input deck
    assumes certain values for any unspecified variables
    If an invalid filepath is given, pre-packaged resources will be checked for
    input decks with a matching name.
    '''
    # TODO generate both an Aviary input file and a phase_info file

    vehicle_data = {'input_values': NamedValues(), 'unused_values': NamedValues(),
                    'initial_guesses': initial_guesses, 'debug_mode': False}

    fortran_deck: Path = get_path(fortran_deck, verbose=False)

    if not out_file:
        name = fortran_deck.stem
        out_file: Path = fortran_deck.parents / name + '_converted.csv'

    if legacy_code is GASP:
        default_extension = '.dat'
        deprecated_vars = gasp_deprecated_vars
    elif legacy_code is FLOPS:
        default_extension = '.txt'
        deprecated_vars = flops_deprecated_vars

    if not defaults_deck:
        defaults_filename = legacy_code.value.lower() + '_default_values' + default_extension
        defaults_deck = Path(__file__).parent.resolve().joinpath(
            'legacy_code_data', defaults_filename)

    # create dictionary to convert legacy code variables to Aviary variables
    aviary_variable_dict = generate_aviary_names([legacy_code.value])

    if defaults_deck:  # If defaults are specified, initialize the vehicle with them
        vehicle_data = input_parser(defaults_deck, vehicle_data,
                                    aviary_variable_dict, deprecated_vars, legacy_code)

    vehicle_data = input_parser(fortran_deck, vehicle_data,
                                aviary_variable_dict, deprecated_vars, legacy_code)
    if legacy_code is GASP:
        vehicle_data = update_gasp_options(vehicle_data)

    if not out_file.is_file():  # default outputted file to be in same directory as input
        out_file = fortran_deck.parent / out_file

    if out_file.is_file():
        if force:
            print(f'Overwriting existing file: {out_file.name}')
        else:
            raise RuntimeError(f'{out_file} already exists. Choose a new name or enable '
                               '--force')
    else:
        # create any directories defined by the new filename if they don't already exist
        out_file.parent.mkdir(parents=True, exist_ok=True)
        print('Writing to:', out_file)

    # open the file in write mode
    with open(out_file, 'w', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(['debug_mode', vehicle_data['debug_mode']])
        writer.writerow([])

        # Values that have been successfully translated to Aviary variables
        writer.writerow(['# Input Values'])
        for var, (val, units) in sorted(vehicle_data['input_values']):
            writer.writerow([var] + val + [units])
        if legacy_code is FLOPS:
            EOM = 'height_energy'
            mass = 'FLOPS'
        if legacy_code is GASP:
            EOM = '2DOF'
            mass = 'GASP'
        writer.writerow(['settings:equations_of_motion'] + [EOM])
        writer.writerow(['settings:mass_method'] + [mass])

        if legacy_code is GASP:
            # Values used in initial guessing of the trajectory
            writer.writerow([])
            writer.writerow(['# Initial Guesses'])
            for var_name in sorted(vehicle_data['initial_guesses']):
                row = [var_name, vehicle_data['initial_guesses'][var_name]]
                writer.writerow(row)

        # Values that were not successfully converted
        writer.writerow([])
        writer.writerow(['# Unconverted Values'])
        for var, (val, _) in sorted(vehicle_data['unused_values']):
            writer.writerow([var] + val)


def input_parser(fortran_deck, vehicle_data, alternate_names, unused_vars, legacy_code):
    '''
    input_parser will modify the values in the vehicle_data dictionary using the data in the
    fortran_deck.
    Lines are read one by one, comments are removed, and namelists are tracked.
    Lines with multiple variable-data pairs are supported, but the last value per variable must
    be followed by a trailing comma.
    '''
    with open(fortran_deck, 'r') as f_in:
        current_namelist = current_tag = ''
        for line in f_in:
            terminate_namelist = False

            tmp = [*line.split('!', 1), '']
            line, comment = tmp[0], tmp[1]  # anything after the first ! is a comment

            # remove all white space and trailing commas
            line = ''.join(line.split()).rstrip(',')
            if len(line.split()) == 0:
                continue  # skip line if it contains only white space

            # Track when namelists are opened and closed
            if (line.lstrip()[0] in ['$', '&']) and current_tag == '':
                current_tag = line.lstrip()[0]
                current_namelist = line.split(current_tag)[1].split()[0] + '.'
            elif (line.lstrip()[0] == current_tag) or (line.rstrip()[-1] == '/'):
                line = line.replace('/', '')
                terminate_namelist = True

            number_of_variables = line.count('=')
            if number_of_variables == 1:
                # get the first element and remove white space
                var_name = ''.join(line.split('=')[0].split())
                # everything after the = is the data
                data = line.split('=')[1]
                try:
                    vehicle_data = process_and_store_data(
                        data, var_name, legacy_code, current_namelist, alternate_names, vehicle_data, unused_vars, comment)
                except Exception as err:
                    if current_namelist == '':
                        raise RuntimeError(line + ' could not be parsed successfully.'
                                           '\nIf this was intended as a comment, '
                                           'add an "!" at the beginning of the line.') from err
                    else:
                        raise err
            elif number_of_variables > 1:
                sub_line = line.split('=')  # split the line at each =
                var_name = sub_line[0]  # the first element is the first name
                for ii in range(number_of_variables):
                    # Each of the following elements contains all of the data for the current variable
                    # and the last element is the name of the next variable
                    sub_list = sub_line[ii+1].split(',')
                    if ii+1 < number_of_variables:
                        next_var_name = sub_list.pop()
                        if not next_var_name[0].isalpha():
                            index = next((i for i, c in enumerate(
                                next_var_name) if c.isalpha()), len(next_var_name))
                            sub_list.append(next_var_name[:index])
                            next_var_name = next_var_name[index:]

                    data = ','.join(sub_list)
                    try:
                        vehicle_data = process_and_store_data(
                            data, var_name, legacy_code, current_namelist, alternate_names, vehicle_data, unused_vars, comment)
                    except Exception as err:
                        if current_namelist == '':
                            raise RuntimeError(line + ' could not be parsed successfully.'
                                               '\nIf this was intended as a comment, '
                                               'add an "!" at the beginning of the line.') from err
                        else:
                            raise err
                    var_name = next_var_name

            if terminate_namelist:
                current_namelist = current_tag = ''

    return vehicle_data


def process_and_store_data(data, var_name, legacy_code, current_namelist, alternate_names, vehicle_data, unused_vars, comment=''):
    '''
    process_and_store_data takes in a string that contains the data, the current variable's name and
    namelist, the dictionary of alternate names, and the current vehicle data.
    It will convert the string of data into a list, get units, check whether the data specified is
    part of a list or a single element, and update the current name to it's equivalent Aviary name.
    The variables are also sorted based on whether they will set an Aviary variable or they are for initial guessing
    '''

    guess_names = list(initial_guesses.keys())
    var_ind = data_units = None
    skip_variable = False
    # skip any variables that shouldn't get converted
    if re.search(current_namelist+var_name+'\Z', str(unused_vars), re.IGNORECASE):
        return vehicle_data
    # remove any elements that are empty (caused by trailing commas or extra commas)
    data_list = [dat for dat in data.split(',') if dat != '']
    if len(data_list) > 0:
        if valid_units(data_list[-1]):
            # if the last element is a unit, remove it from the list and update the variable's units
            data_units = data_list.pop()
        var_values = convert_strings_to_data(data_list)
    else:
        skip_variable = True
        var_values = []

    if '(' in var_name:  # some GASP lists are given as individual elements
        # get the target index (Fortran uses 1 indexing, Python uses 0 indexing)
        fortran_offset = 1 if current_namelist else 0
        var_ind = int(var_name.split('(')[1].split(')')[0])-fortran_offset
        var_name = var_name.split('(')[0]  # remove the index formatting

    list_of_equivalent_aviary_names = update_name(
        alternate_names, current_namelist+var_name, vehicle_data['debug_mode'])
    for name in list_of_equivalent_aviary_names:
        if not skip_variable:
            if name == 'debug_mode':
                vehicle_data['debug_mode'] = var_values[0]
                continue

            elif name in guess_names and legacy_code is GASP:
                # all initial guesses take only a single value
                vehicle_data['initial_guesses'][name] = float(var_values[0])
                continue

            elif name in _MetaData:
                vehicle_data['input_values'] = set_value(name, var_values, vehicle_data['input_values'],
                                                         var_ind=var_ind, units=data_units)
                continue

        vehicle_data['unused_values'] = set_value(name, var_values, vehicle_data['unused_values'],
                                                  var_ind=var_ind, units=data_units)
        if vehicle_data['debug_mode']:
            print('Unused:', name, var_values, comment)

    return vehicle_data


def set_value(var_name, var_value, value_dict: NamedValues, var_ind=None, units=None):
    ''' 
    set_value will update the current value of a variable in a value dictionary that contains a value
    and it's associated units.
    If units are specified for the new value, they will be used, otherwise the current units in the
    value dictionary or the default units from _MetaData are used.
    If the new variable is part of a list, the current list will be extended if needed.
    '''

    if var_name in value_dict:
        current_value, units = value_dict.get_item(var_name)
    else:
        current_value = None
        if var_name in _MetaData:
            units = _MetaData[var_name]['units']
        else:
            units = 'unitless'
    if not units:
        units = 'unitless'

    if var_ind != None:
        # if an index is specified, use it, otherwise store the input as the whole value
        if isinstance(current_value, list):
            max_ind = len(current_value) - 1
            if var_ind > max_ind:
                current_value.extend((var_ind-max_ind)*[0])
        else:
            current_value = [current_value]+[0]*var_ind
        current_value[var_ind] = var_value[0]
        value_dict.set_val(var_name, current_value, units)
    else:
        value_dict.set_val(var_name, var_value, units)
    return value_dict


def generate_aviary_names(code_bases):
    '''
    Create a dictionary for each of the specified Fortran code bases to map to the Aviary
    variable names. Each dictionary of Aviary names will have a list of Fortran names for
    each variable
    '''

    alternate_names = {}
    for code_base in code_bases:
        alternate_names[code_base] = {}
        for key in _MetaData.keys():
            historical_dict = _MetaData[key]['historical_name']
            if historical_dict and code_base in historical_dict:
                alt_name = _MetaData[key]['historical_name'][code_base]
                if isinstance(alt_name, str):
                    alt_name = [alt_name]
                alternate_names[code_base][key] = alt_name
    return alternate_names


def update_name(alternate_names, var_name, debug_mode=False):
    '''update_name will convert a Fortran name to a list of equivalent Aviary names.'''

    all_equivalent_names = []
    for code_base in alternate_names.keys():
        for key, list_of_names in alternate_names[code_base].items():
            if list_of_names is not None:
                if any([re.search(var_name+r'\Z', altname, re.IGNORECASE) for altname in list_of_names]):
                    all_equivalent_names.append(key)

    # if there are no equivalent variable names, return the original name
    if len(all_equivalent_names) == 0:
        if debug_mode:
            print('passing: ', var_name)
        all_equivalent_names = [var_name]
    return all_equivalent_names


def update_gasp_options(vehicle_data):
    """
    Handles variables that are affected by the values of others
    """
    input_values: NamedValues = vehicle_data['input_values']

    flap_types = ["plain", "split", "single_slotted", "double_slotted",
                  "triple_slotted", "fowler", "double_slotted_fowler"]

    ## PROBLEM TYPE ##
    # if multiple values of target_range are specified, use the one that corresponds to the problem_type
    design_range, distance_units = input_values.get_item(Mission.Design.RANGE)
    try:
        problem_type = input_values.get_val('problem_type')[0]
    except KeyError:
        problem_type = 'sizing'

    if isinstance(design_range, list):
        # if the design range target_range value is 0, set the problem_type to fallout
        if design_range[0] == 0:
            input_values.set_val('problem_type', ['fallout'])
            design_range = 0
        if problem_type == 'sizing':
            design_range = design_range[0]
        elif problem_type == 'alternate':
            design_range = design_range[2]
        elif problem_type == 'fallout':
            design_range = 0
    else:
        if design_range == 0:
            input_values.set_val('problem_type', ['fallout'])
    input_values.set_val(Mission.Design.RANGE, [design_range], distance_units)

    ## STRUT AND FOLD ##
    strut_loc = input_values.get_val(Aircraft.Strut.ATTACHMENT_LOCATION, 'ft')[0]
    folded_span = input_values.get_val(Aircraft.Wing.FOLDED_SPAN, 'ft')[0]

    if strut_loc == 0:
        input_values.set_val(Aircraft.Wing.HAS_STRUT, [False], 'unitless')
    else:
        input_values.set_val(Aircraft.Wing.HAS_STRUT, [True], 'unitless')
    if folded_span == 0:
        input_values.set_val(Aircraft.Wing.HAS_FOLD, [False], 'unitless')
    else:
        input_values.set_val(Aircraft.Wing.HAS_FOLD, [True], 'unitless')

    if strut_loc < 0:
        input_values.set_val(Aircraft.Wing.HAS_FOLD, [True], 'unitless')
        input_values.set_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, [False], 'unitless')
        strut_loc = abs(strut_loc)

    if strut_loc < 1:
        input_values.set_val(Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS,
                             [strut_loc], 'unitless')
        input_values.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, [
                             False], 'unitless')
    else:
        input_values.set_val(Aircraft.Strut.ATTACHMENT_LOCATION, [strut_loc], 'ft')
        input_values.set_val(
            Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, [True], 'unitless')

    if input_values.get_val(Aircraft.Wing.HAS_FOLD)[0]:
        if not input_values.get_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION)[0]:
            input_values.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                                 [True], 'unitless')
        else:
            if input_values.get_val(Aircraft.Wing.FOLDED_SPAN, 'ft')[0] > 1:
                input_values.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                                     [True], 'unitless')
            else:
                input_values.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                                     [False], 'unitless')
    else:
        input_values.set_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, [True], 'unitless')
        input_values.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, [
                             False], 'unitless')

    ## FLAPS ##
    flap_type = input_values.get_val(Aircraft.Wing.FLAP_TYPE)[0]
    if not isinstance(flap_type, str):
        flap_type = flap_types[flap_type-1]
        input_values.set_val(Aircraft.Wing.FLAP_TYPE, [flap_type])
    flap_ind = flap_types.index(flap_type)
    if input_values.get_val(Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT)[0] <= 0:
        input_values.set_val(Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT,
                             [[0.62, 1.0, 0.733, 1.2, 1.32, 0.633, 0.678][flap_ind]])
    if input_values.get_val(Aircraft.Wing.OPTIMUM_FLAP_DEFLECTION, 'deg')[0] == 0:
        input_values.set_val(Aircraft.Wing.OPTIMUM_FLAP_DEFLECTION,
                             [[60, 60, 40, 55, 55, 30, 30][flap_ind]], 'deg')
    if input_values.get_val(Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM)[0] == 0:
        input_values.set_val(Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM,
                             [[.9, .8, 1.18, 1.4, 1.6, 1.67, 2.25][flap_ind]])
    if input_values.get_val(Aircraft.Wing.FLAP_DRAG_INCREMENT_OPTIMUM)[0] == 0:
        input_values.set_val(Aircraft.Wing.FLAP_DRAG_INCREMENT_OPTIMUM,
                             [[.12, .23, .13, .23, .23, .1, .15][flap_ind]])

    res = input_values.get_val(Aircraft.Design.RESERVE_FUEL_ADDITIONAL, units='lbm')[0]
    if res <= 0:
        input_values.set_val(Aircraft.Design.RESERVE_FUEL_ADDITIONAL, [0], units='lbm')
        input_values.set_val(Aircraft.Design.RESERVE_FUEL_FRACTION,
                             [-res], units='unitless')
    elif res >= 10:
        input_values.set_val(Aircraft.Design.RESERVE_FUEL_FRACTION,
                             [0], units='unitless')
    else:
        ValueError('"FRESF" is not valid between 0 and 10.')

    vehicle_data['input_values'] = input_values
    return vehicle_data


def update_flops_options(vehicle_data):
    """
    Handles variables that are affected by the values of others
    """
    input_values: NamedValues = vehicle_data['input_values']

    for var_name in flops_scalar_variables.items():
        update_flops_scalar_variables(var_name, input_values)

    # TWR <= 0 is not valid in Aviary (parametric variation)
    if Aircraft.Design.THRUST_TO_WEIGHT_RATIO in input_values:
        if input_values.get_val(Aircraft.Design.THRUST_TO_WEIGHT_RATIO) <= 0:
            input_values.delete(Aircraft.Design.THRUST_TO_WEIGHT_RATIO)

    # WSR

    # Additional mass fraction scalar set to zero to not add mass twice
    if Aircraft.Engine.ADDITIONAL_MASS_FRACTION in input_values:
        if input_values.get_val(Aircraft.Engine.ADDITIONAL_MASS_FRACTION) >= 1:
            input_values.set_val(Aircraft.Engine.ADDITIONAL_MASS,
                                 input_values.get_val(
                                     Aircraft.Engine.ADDITIONAL_MASS_FRACTION),
                                 'lbm')
            input_values.set_val(Aircraft.Engine.ADDITIONAL_MASS_FRACTION, 0.0)

    # Miscellaneous propulsion mass trigger point 1 instead of 5
    if Aircraft.Propulsion.MISC_MASS_SCALER in input_values:
        if input_values.get_val(Aircraft.Propulsion.MISC_MASS_SCALER) >= 1:
            input_values.set_val(Aircraft.Propulsion.TOTAL_MISC_MASS,
                                 input_values.get_val(
                                     Aircraft.Propulsion.MISC_MASS_SCALER),
                                 'lbm')
            input_values.set_val(Aircraft.Propulsion.MISC_MASS_SCALER, 0.0)

    vehicle_data['input_values'] = input_values
    return vehicle_data


def update_flops_scalar_variables(var_name, input_values: NamedValues):
    # The following parameters are used to modify or override
    # internally computed weights for various components as follows:
    # < 0., negative of starting weight which will be modified
    #   as appropriate during optimization or parametric
    #   variation, lb
    # = 0., no weight for that component
    # > 0. but < 5., scale factor applied to internally
    #   computed weight
    # > 5., actual fixed weight for component, lb
    # Same rules also applied to various other FLOPS scalar parameters
    scalar_name = var_name + '_scaler'
    if scalar_name not in input_values:
        return
    scalar_value = input_values[scalar_name]
    if scalar_value <= 0:
        input_values.delete(scalar_name)
    elif scalar_value < 5:
        return
    elif scalar_value > 5:
        input_values.set_val(var_name, scalar_value, 'lbm')
        input_values.set_val(scalar_name, 1.0)


# list storing information on Aviary variables that are split from single
# FLOPS variables that use the same value-based branching behavior
flops_scalar_variables = [
    Aircraft.AirConditioning.MASS,
    Aircraft.AntiIcing.MASS,
    Aircraft.APU.MASS,
    Aircraft.Avionics.MASS,
    Aircraft.Canard.MASS,
    Aircraft.Canard.WETTED_AREA,
    Aircraft.CrewPayload.CARGO_CONTAINER_MASS,
    Aircraft.CrewPayload.FLIGHT_CREW_MASS,
    Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS,
    Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
    Aircraft.Design.EMPTY_MASS_MARGIN,
    Aircraft.Electrical.MASS,
    Aircraft.Engine.THRUST_REVERSERS_MASS,
    Aircraft.Fins.MASS,
    Aircraft.Fuel.FUEL_SYSTEM_MASS,
    Aircraft.Fuel.UNUSABLE_FUEL_MASS,
    Aircraft.Furnishings.MASS,
    Aircraft.Fuselage.MASS,
    Aircraft.Fuselage.WETTED_AREA,
    Aircraft.HorizontalTail.MASS,
    Aircraft.HorizontalTail.WETTED_AREA,
    Aircraft.Hydraulics.MASS,
    Aircraft.Instruments.MASS,
    Aircraft.LandingGear.MAIN_GEAR_MASS,
    Aircraft.LandingGear.NOSE_GEAR_MASS,
    Aircraft.Nacelle.MASS,
    Aircraft.Propulsion.TOTAL_ENGINE_OIL_MASS,
    Aircraft.VerticalTail.MASS_SCALER,
    Aircraft.VerticalTail.WETTED_AREA_SCALER,
    Aircraft.Wing.MASS,
    Aircraft.Wing.SHEAR_CONTROL_MASS,
    Aircraft.Wing.SURFACE_CONTROL_MASS,
    Aircraft.Wing.WETTED_AREA,
]

initial_guesses = {
    # initial_guesses is a dictionary that contains values used to initialize the trajectory
    'actual_takeoff_mass': 0,
    'rotation_mass': .99,
    'fuel_burn_per_passenger_mile': 0.1,
    'cruise_mass_final': 0,
    'flight_duration': 0,
    'time_to_climb': 0,
    'climb_range': 0,
    'reserves': 0
}


def _setup_F2A_parser(parser):
    '''
    Set up the subparser for the Fortran_to_aviary tool.

    Parameters
    ----------
    parser : argparse subparser
        The parser we're adding options to.
    '''
    parser.add_argument(
        "input_deck",
        type=str,
        nargs=1,
        help="Filename of vehicle input deck, including partial or complete path.",
    )
    parser.add_argument(
        "-o",
        "--out_file",
        default=None,
        help="Filename for converted input deck, including partial or complete path."
    )
    parser.add_argument(
        "-l",
        "--legacy_code",
        type=LegacyCode,
        help="Name of the legacy code the deck originated from",
        choices=list(LegacyCode),
        required=True
    )
    parser.add_argument(
        "-d",
        "--defaults_deck",
        default=None,
        help="Deck of default values for unspecified variables"
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Allow overwriting existing output files",
    )


def _exec_F2A(args, user_args):
    # check if args.input_deck is a list, if so, use the first element
    if isinstance(args.input_deck, list):
        args.input_deck = args.input_deck[0]
    filepath = args.input_deck

    create_aviary_deck(filepath, args.legacy_code, args.defaults_deck,
                       Path(args.out_file), args.force)



import openmdao.api as om

from aviary.subsystems.mass.flops_based.landing_gear import (
    AltLandingGearMass, LandingGearMass, MainGearLength, NoseGearLength)
from aviary.subsystems.mass.flops_based.landing_mass import (
    LandingMass, LandingTakeoffMassRatio)
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Mission


class LandingMassGroup(om.Group):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        aviary_options: AviaryValues = self.options['aviary_options']
        alt_mass = aviary_options.get_val(Aircraft.Design.USE_ALT_MASS)

        self.add_subsystem('landing_to_takeoff_mass_ratio',
                           LandingTakeoffMassRatio(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('main_landing_gear_length',
                           MainGearLength(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('nose_landing_gear_length',
                           NoseGearLength(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('landing_mass',
                           LandingMass(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        if alt_mass:
            self.add_subsystem('landing_gear',
                               AltLandingGearMass(aviary_options=aviary_options),
                               promotes_inputs=['*'], promotes_outputs=['*'])
        else:
            self.add_subsystem('landing_gear',
                               LandingGearMass(aviary_options=aviary_options),
                               promotes_inputs=['*'], promotes_outputs=['*'])
